"""A FIFO queue for processing textual Paradict data"""
from collections import namedtuple


class TxtQueue:
    """A FIFO queue for processing textual Paradict data"""
    def __init__(self):
        self._buffer = list()

    @property
    def buffer(self):
        return self._buffer

    def enqueue(self, s):
        """Store textual data in the buffer. This data will then be iteratively
        extracted by the 'get' method, line by line.
        Note: make sure that each line is ended with a newline '\n' character"""
        self._buffer.append(s)

    def dequeue(self):
        """Generator for iteratively getting each line composing the
        textual data stored in the buffer.
        Note that lines yielded won't end with a newline '\n' character"""
        yield from self._read()

    def _read(self):
        if not self._buffer:
            return
        text = "".join(self._buffer)
        del self._buffer[:]
        cache = list()
        i = None
        for i, char in enumerate(text):
            cache.append(char)
            if char == "\n":
                line = "".join(cache)
                del cache[:]
                yield line
        if cache and i is not None:
            self._buffer.insert(0, "".join(cache))



import unittest
import datetime
from decimal import Decimal
from paradict import validator, box
from paradict.errors import Error
from paradict.typeref import TypeRef


class TestDataValidation(unittest.TestCase):

    def test_string(self):
        data = "hello world"
        schema = "str"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        data = ''
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_int(self):
        data = 1234567
        schema = "int"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        data = box.HexInt(1234567)
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_float(self):
        data = 1.234567
        schema = "float"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        data = Decimal("1.234567")
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_bin(self):
        data = b"hello world"
        schema = "bin"
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_bool(self):
        data = True
        schema = "bool"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        data = False
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_complex(self):
        data = 1+2j
        schema = "complex"
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_date(self):
        data = datetime.date.today()
        schema = "date"
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_time(self):
        data = datetime.time(10, 30, 58)
        schema = "time"
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_datetime(self):
        data = datetime.datetime.now()
        schema = "datetime"
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_grid(self):
        data = box.Grid()
        schema = "grid"
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestDataValidationWithSpec(unittest.TestCase):

    def test_string(self):
        data = "hello world"
        schema = validator.Spec("str")
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        data = "hello"
        schema = validator.Spec("str", lambda x: len(x) == 5)
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        data = ''
        schema = validator.Spec("str", lambda x: len(x) == 0)
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_int(self):
        data = 42
        schema = validator.Spec("int")
        r = validator.validate(data, schema)
        self.assertTrue(r)
        #
        schema = validator.Spec("int", lambda x: x > 40 and x < 45)
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestDataInvalidation(unittest.TestCase):

    def test_string(self):
        data = "hello world"
        schema = "int"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_int(self):
        data = 1234567
        schema = "str"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_float(self):
        data = 1.234567
        schema = "int"
        r = validator.validate(data, schema)
        self.assertFalse(r)
        #
        data = Decimal("1.234567")
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_bin(self):
        data = b"hello world"
        schema = "str"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_bool(self):
        data = True
        schema = "int"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_complex(self):
        data = 1+2j
        schema = "int"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_date(self):
        data = datetime.date.today()
        schema = "datetime"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_time(self):
        data = datetime.time(10, 30, 58)
        schema = "datetime"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_datetime(self):
        data = datetime.datetime.now()
        schema = "date"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_grid(self):
        data = list()
        schema = "grid"
        r = validator.validate(data, schema)
        self.assertFalse(r)


class TestDataInvalidationWithSpec(unittest.TestCase):

    def test_string(self):
        data = "hello world"
        schema = validator.Spec("int")
        r = validator.validate(data, schema)
        self.assertFalse(r)
        #
        data = "hello"
        schema = validator.Spec("str", lambda x: len(x) != 5)
        r = validator.validate(data, schema)
        self.assertFalse(r)
        #
        data = ''
        schema = validator.Spec("str", lambda x: len(x) != 0)
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_int(self):
        data = 42
        schema = "str"
        r = validator.validate(data, schema)
        self.assertFalse(r)
        #
        schema = validator.Spec("str")
        r = validator.validate(data, schema)
        self.assertFalse(r)
        #
        schema = validator.Spec("int", lambda x: x < 40)
        r = validator.validate(data, schema)
        self.assertFalse(r)


class TestDictValidation(unittest.TestCase):

    def test_non_empty_dict(self):
        data = {"name": "alex", "age": 420}
        # test 1
        schema = "dict"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # test 2
        schema = {"name": "str", "age": "int"}
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # test 3
        data = {"name": "alex"}
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_nested_non_empty_dict(self):
        data = {"name": "alex", "age": 420,
                "books": {"sci-fi":
                              ["book1", "book2"],
                          "thriller":
                              ["book3", "book4"]}}
        # test 1
        schema = {"name": "str", "age": "int",
                  "books": {"sci-fi":
                                ["str"],
                            "thriller":
                                ["str"]}}
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestDictInvalidation(unittest.TestCase):

    def test_type_mismatch(self):
        data = dict()
        schema = "list"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_non_empty_dict(self):
        data = {"name": "alex", "age": 420}
        # test 1
        schema = {"name": "int", "age": "int"}
        r = validator.validate(data, schema)
        self.assertFalse(r)
        # test 2
        schema = "list"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_nested_non_empty_dict(self):
        data = {"name": "alex", "age": 420,
                "books": {"sci-fi": "book1",
                          "thriller": "book2"}}
        # test 1
        schema = {"name": "str", "age": "int",
                  "books": {"sci-fi": "str",
                            "thriller": "int"}}
        r = validator.validate(data, schema)
        self.assertFalse(r)


class TestEmptyDict(unittest.TestCase):

    def test_empty_data(self):
        data = dict()
        schema = "dict"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # ---
        schema = dict()
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # ---
        schema = {"name": "str", "age": "int"}
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_empty_schema(self):
        # ---
        data = {"name": "alex"}
        schema = dict()
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestListValidation(unittest.TestCase):

    def test_list_of_int(self):
        data = [2, 3, 5, 7, 11]
        # test 1
        schema = ["int"]
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # test 2
        schema = "list"
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_list_of_int_and_str(self):
        # test 1
        data = [1, "a", 2, "b", None]
        schema = ["int", "str"]
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_nested_list_of_int(self):
        # test 1
        data = [1, 2, 3, [4, 5, 6, [7, 8, 9]]]
        schema = ["int", ["int", ["int"]]]
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # test 2
        data = [1, 2, 3]
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_nested_list_of_int_and_str(self):
        # test 1
        data = [1, 2, 3, ["a", "b", "c", [4, 5, 6]]]
        schema = ["int", ["str", ["int"]]]
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestListInvalidation(unittest.TestCase):

    def test_type_mismatch(self):
        data = list()
        schema = "dict"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_list_of_int(self):
        data = [2, 3, 5, 7, 11]
        schema = ["str"]
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_list_of_int_and_str(self):
        data = [1, "a", 2, "b", None]
        # test 1
        schema = ["str"]
        r = validator.validate(data, schema)
        self.assertFalse(r)
        # test 2
        schema = ["int"]
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_nested_list_of_int(self):
        data = [1, 2, 3, [4, 5, 6, [7, 8, 9]]]
        # test 1
        schema = ["int"]
        r = validator.validate(data, schema)
        self.assertFalse(r)
        # test 2
        schema = ["int", ["int"]]
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_nested_list_of_int_and_str(self):
        data = [1, 2, 3, ["a", "b", "c", [4, 5, 6]]]
        # test 1
        schema = ["int", ["str"]]
        r = validator.validate(data, schema)
        self.assertFalse(r)
        # test 2
        schema = ["int", "str", "int"]
        r = validator.validate(data, schema)
        self.assertFalse(r)


class TestEmptyList(unittest.TestCase):

    def test_empty_data(self):
        data = list()
        schema = "list"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # ---
        schema = list()
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # ---
        schema = ["int"]
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_empty_schema(self):
        data = ["alex", "rustic"]
        schema = list()
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestObjValidation(unittest.TestCase):

    def test_non_empty_obj(self):
        data = box.Obj({"name": "alex", "age": 420})
        # test 1
        schema = "obj"
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # test 2
        schema = box.Obj({"name": "str", "age": "int"})
        r = validator.validate(data, schema)
        self.assertTrue(r)
        # test 3
        data = box.Obj({"name": "alex"})
        r = validator.validate(data, schema)
        self.assertTrue(r)

    def test_nested_non_empty_obj(self):
        data = box.Obj({"name": "alex", "age": 420,
                        "books": {"sci-fi":
                                      ["book1", "book2"],
                                  "thriller":
                                      ["book3", "book4"]}})
        # test 1
        schema = box.Obj({"name": "str", "age": "int",
                          "books": {"sci-fi":
                                        ["str"],
                                    "thriller":
                                        ["str"]}})
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestObjInvalidation(unittest.TestCase):

    def test_type_mismatch(self):
        data = box.Obj()
        schema = "list"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_non_empty_obj(self):
        data = box.Obj({"name": "alex", "age": 420})
        # test 1
        schema = box.Obj({"name": "int", "age": "int"})
        r = validator.validate(data, schema)
        self.assertFalse(r)
        # test 2
        schema = "list"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_nested_non_empty_obj(self):
        data = box.Obj({"name": "alex", "age": 420,
                        "books": {"sci-fi": "book1",
                                  "thriller": "book2"}})
        # test 1
        schema = box.Obj({"name": "str", "age": "int",
                          "books": {"sci-fi": "str",
                                    "thriller": "int"}})
        r = validator.validate(data, schema)
        self.assertFalse(r)


class TestComplexStructure(unittest.TestCase):

    def test(self):
        data = {"name": "alex", "age": 420,
                "books": {"sci-fi":
                              ["book1", "book2"],
                          "thriller":
                              ["book3", "book4"]},
                "misc": [1, 3.14, "pi"],
                "empty": None,
                "n": 42}
        # test 1
        schema = {"name": "str", "age": "int",
                  "books": {"sci-fi":
                                ["str"],
                            "thriller":
                                ["str"]},
                  "misc": ["int", "float", "str"],
                  "empty": "str",
                  "n": validator.Spec("int", lambda x: x > 40 and x < 45)}
        r = validator.validate(data, schema)
        self.assertTrue(r)


class TestInvalidSchema(unittest.TestCase):

    def test_invalidation(self):
        data = "hello world"
        schema = ["string"]
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_exception_raise(self):
        data = ["hello world"]
        schema = ["string"]
        with self.assertRaises(Error):
            validator.validate(data, schema)


class TestIrregularData(unittest.TestCase):

    def test_invalidation(self):
        data = ExoticData()
        schema = "str"
        r = validator.validate(data, schema)
        self.assertFalse(r)


class TestTypeRef(unittest.TestCase):

    def test_invalidation(self):
        data = CustomInteger()
        schema = "int"
        r = validator.validate(data, schema)
        self.assertFalse(r)

    def test_validation(self):
        data = CustomInteger()
        schema = "int"
        type_ref = TypeRef()
        type_ref.int_types.append(CustomInteger)
        r = validator.validate(data, schema, type_ref=type_ref)
        self.assertTrue(r)


class ExoticData:
    pass


class CustomInteger(int):
    pass


if __name__ == '__main__':
    unittest.main()



"""A FIFO queue for processing binary Paradict data"""
from collections import namedtuple
from paradict import tags, errors
from paradict.tags.misc import SIZE_TO_STR
from paradict import misc

# a datum is made of a header and payload
# The header is made of a tag and optional bytes to store the size of the payload
# The width is the size of the datum: size of header + size of payload
Datum = namedtuple("Datum", ["tag", "payload", "width"])


class BinQueue:
    """A FIFO queue for processing binary Paradict data"""
    def __init__(self):
        self._tag = None
        self._buffer = bytearray()
        self._expected_width = 0

    @property
    def buffer(self):
        return self._buffer

    def enqueue(self, raw):
        """Store binary data in the buffer. This data will then be iteratively
        extracted by the 'get' method"""
        self._buffer.extend(raw)

    def dequeue(self):
        """Generator for iteratively getting each tag-payload tuple composing the
        raw data stored in the buffer"""
        while True:
            r = self._read()
            if not r:
                break
            tag, payload = r
            yield tag, payload

    def _read(self):
        if not self._buffer:
            return
        # update tag
        self._update_tag_var()
        # update span
        self._update_expected_width_var()
        if len(self._buffer) < self._expected_width:
            return
        # STRx
        if tags.STR_8 <= self._tag <= tags.STR_256:
            return self._read_buffer(read_strx)
        # STR
        elif self._tag in (tags.STR_SHORT,
                           tags.STR_MEDIUM,
                           tags.STR_LONG,
                           tags.STR_BIG,
                           tags.STR_HEAVY):
            return self._read_buffer(read_str)
        # BIN
        elif self._tag in (tags.BIN_SHORT,
                           tags.BIN_MEDIUM,
                           tags.BIN_LONG,
                           tags.BIN_BIG,
                           tags.BIN_HEAVY):
            return self._read_buffer(read_bin)
        # PINTx
        elif tags.PINT_8 <= self._tag <= tags.PINT_64:
            return self._read_buffer(read_pintx)
        # PINT_BIG
        elif self._tag == tags.PINT_BIG:
            return self._read_buffer(read_pint_big)
        # PINT_HEAVY
        elif self._tag == tags.PINT_HEAVY:
            return self._read_buffer(read_pint_heavy)
        # NINTx
        elif tags.NINT_8 <= self._tag <= tags.NINT_64:
            return self._read_buffer(read_nintx)
        # NINT_BIG
        elif self._tag == tags.NINT_BIG:
            return self._read_buffer(read_nint_big)
        # NINT_HEAVY
        elif self._tag == tags.NINT_HEAVY:
            return self._read_buffer(read_nint_heavy)
        # else...
        else:
            return self._read_buffer(read_tag)

    def _read_buffer(self, reader):
        r = reader(self._tag, self._buffer)
        if r is None:
            return
        if r.payload is None:
            self._expected_width = r.width
            return
        self._buffer[:] = self._buffer[r.width:]
        self._clear_status()
        return r.tag, r.payload

    def _update_tag_var(self):
        if self._tag is None:
            self._tag = self._buffer[0].to_bytes(length=1,
                                                 byteorder="little",
                                                 signed=False)

    def _update_expected_width_var(self):
        if self._expected_width or self._tag is None:
            return
        if self._tag in (tags.PINT_8, tags.NINT_8, tags.STR_8):
            self._expected_width = 2
        elif self._tag in (tags.PINT_16, tags.NINT_16, tags.STR_16):
            self._expected_width = 3
        elif self._tag in (tags.PINT_24, tags.NINT_24, tags.STR_24):
            self._expected_width = 4
        elif self._tag in (tags.PINT_32, tags.NINT_32, tags.STR_32):
            self._expected_width = 5
        elif self._tag in (tags.PINT_40, tags.NINT_40, tags.STR_40):
            self._expected_width = 6
        elif self._tag in (tags.PINT_48, tags.NINT_48, tags.STR_48):
            self._expected_width = 7
        elif self._tag in (tags.PINT_56, tags.NINT_56, tags.STR_56):
            self._expected_width = 8
        elif self._tag in (tags.PINT_64, tags.NINT_64, tags.STR_64):
            self._expected_width = 9
        elif self._tag == tags.STR_72:
            self._expected_width = 10
        elif self._tag == tags.STR_80:
            self._expected_width = 11
        elif self._tag == tags.STR_88:
            self._expected_width = 12
        elif self._tag == tags.STR_96:
            self._expected_width = 13
        elif self._tag == tags.STR_104:
            self._expected_width = 14
        elif self._tag == tags.STR_112:
            self._expected_width = 15
        elif self._tag == tags.STR_120:
            self._expected_width = 16
        elif self._tag == tags.STR_128:
            self._expected_width = 17
        elif self._tag == tags.STR_136:
            self._expected_width = 18
        elif self._tag == tags.STR_144:
            self._expected_width = 19
        elif self._tag == tags.STR_152:
            self._expected_width = 20
        elif self._tag == tags.STR_160:
            self._expected_width = 21
        elif self._tag == tags.STR_168:
            self._expected_width = 22
        elif self._tag == tags.STR_176:
            self._expected_width = 23
        elif self._tag == tags.STR_184:
            self._expected_width = 24
        elif self._tag == tags.STR_192:
            self._expected_width = 25
        elif self._tag == tags.STR_200:
            self._expected_width = 26
        elif self._tag == tags.STR_208:
            self._expected_width = 27
        elif self._tag == tags.STR_216:
            self._expected_width = 28
        elif self._tag == tags.STR_224:
            self._expected_width = 29
        elif self._tag == tags.STR_232:
            self._expected_width = 30
        elif self._tag == tags.STR_240:
            self._expected_width = 31
        elif self._tag == tags.STR_248:
            self._expected_width = 32
        elif self._tag == tags.STR_256:
            self._expected_width = 33
        elif self._tag in (tags.STR_SHORT, tags.BIN_SHORT):
            self._expected_width = 3
        elif self._tag in (tags.STR_MEDIUM, tags.BIN_MEDIUM):
            self._expected_width = 4
        elif self._tag in (tags.STR_LONG, tags.BIN_LONG):
            self._expected_width = 5
        elif self._tag in (tags.STR_BIG, tags.BIN_BIG):
            self._expected_width = 6
        elif self._tag in (tags.STR_HEAVY, tags.BIN_HEAVY):
            self._expected_width = 7
        else:
            self._expected_width = 1

    def _clear_status(self):
        self._tag = None
        self._expected_width = 0


def read_payload_size(buffer, nb=1, offset=1):
    """
    Read the expected size of payload
    Note that nb is the number of bytes encoding the size of payload
    """
    r = buffer[offset: offset+nb]
    return int.from_bytes(r, "little", signed=False)


def extract_datum(buffer, tag, nb=1):
    """
    This function is used read STR_SHORT, STR_MEDIUM, STR_LONG, STR_BIG,
     STR_HEAVY, BIN_SHORT, BIN_MEDIUM, BIN_LONG, BIN_BIG, and BIN_HEAVY datums.

    nb is the number of bytes to encode the size of payload

    This function returns a Datum
    """
    # header_size
    header_size = 1 + nb
    if len(buffer) < header_size:
        return
    payload_size = read_payload_size(buffer, nb) + 1
    # Width is the size (bytes) of the datum (header + payload)
    # note that the header is made of tag + byte(s) to store payload size
    width = header_size + payload_size
    if len(buffer) < width:
        payload = None
    else:
        payload = buffer[header_size:width]
    return Datum(tag, payload, width)


def read_tag(tag, buffer):
    payload = b''
    width = 1
    return Datum(tag, payload, width)


def read_str(tag, buffer):
    containers = {tags.STR_SHORT: 1,
                  tags.STR_MEDIUM: 2,
                  tags.STR_LONG: 3,
                  tags.STR_BIG: 4,
                  tags.STR_HEAVY: 5}
    # nb is the number of bytes to encode the size of payload
    nb = containers.get(tag)
    return extract_datum(buffer, tag, nb)


def read_strx(tag, buffer):
    size = None
    for s, t in SIZE_TO_STR.items():
        if t == tag:
            size = s
            break
    if size is None:
        raise errors.Error("Unknown string tag '{}'".format(tag))
    width = 1 + size
    if len(buffer) < width:
        payload = None
    else:
        payload = buffer[1:width]
    return Datum(tag, payload, width)


def read_bin(tag, buffer):
    containers = {tags.BIN_SHORT: 1,
                  tags.BIN_MEDIUM: 2,
                  tags.BIN_LONG: 3,
                  tags.BIN_BIG: 4,
                  tags.BIN_HEAVY: 5}
    # nb is the number of bytes to encode the size of payload
    nb = containers.get(tag)
    return extract_datum(buffer, tag, nb)


def read_pint_big(tag, buffer):
    n = len(buffer)
    if n < 2:  # here, 2 == 1 byte tag + 1 byte to store the size of payload
        return
    payload_size = read_payload_size(buffer) + 1
    width = 2 + payload_size
    if len(buffer) < width:
        payload = None
    else:
        payload = buffer[2:width]
    return Datum(tag, payload, width)


def read_pint_heavy(tag, buffer):
    n = len(buffer)
    if n < 3:  # here, 3 == 1 byte tag + 2 byte to store the size of payload
        return
    payload_size = read_payload_size(buffer, nb=2) + 1
    width = 3 + payload_size
    if len(buffer) < width:
        payload = None
    else:
        payload = buffer[3:width]
    return Datum(tag, payload, width)


def read_pintx(tag, buffer):
    sizes = {tags.PINT_8: 1, tags.PINT_16: 2,
             tags.PINT_24: 3, tags.PINT_32: 4,
             tags.PINT_40: 5, tags.PINT_48: 6,
             tags.PINT_56: 7, tags.PINT_64: 8}
    payload_size = sizes.get(tag)
    width = 1 + payload_size
    if len(buffer) < width:
        payload = None
    else:
        payload = buffer[1:width]
    return Datum(tag, payload, width)


def read_nint_big(tag, buffer):
    return read_pint_big(tag, buffer)


def read_nint_heavy(tag, buffer):
    return read_pint_heavy(tag, buffer)


def read_nintx(tag, buffer):
    sizes = {tags.NINT_8: 1, tags.NINT_16: 2,
             tags.NINT_24: 3, tags.NINT_32: 4,
             tags.NINT_40: 5, tags.NINT_48: 6,
             tags.NINT_56: 7, tags.NINT_64: 8}
    payload_size = sizes.get(tag)
    width = 1 + payload_size
    if len(buffer) < width:
        payload = None
    else:
        payload = buffer[1:width]
    return Datum(tag, payload, width)



import datetime
from paradict import tags
from paradict.tags.misc import ALPHABET
from paradict.typeref import TypeRef
from paradict import errors
from paradict.queue.bin_queue import BinQueue
from paradict import misc, box


class Unpacker:
    """Class to convert some binary Paradict data into a Python dict"""

    def __init__(self, type_ref=None, receiver=None, obj_builder=None,
                 skip_comments=False):
        """
        Init

        [parameters]
        - type_ref: optional TypeRef object
        - receiver: callback function that will be called at the end of conversion.
        This callback function accepts the Unpacker instance as argument
        - obj_builder: function that accepts a paradict.box.Obj container and
        returns a fresh new Python object
        - skip_comments: boolean to tell whether comments should be ignored or not
        """
        self._type_ref = type_ref if type_ref else TypeRef()
        self._receiver = receiver
        self._obj_builder = obj_builder
        self._skip_comments = skip_comments
        self._buffer = bytearray()
        self._index = 0
        self._queue = BinQueue()
        self._feedable = True
        self._stack = list()
        self._data = None
        self._block_on = False
        self._block_count = 0
        # flag
        self._as_comment_str = False
        # byte count for exception message
        self._byte_count = 0

    @property
    def data(self):
        if self._stack:
            context = self._stack[0]
            return context.container
        return self._data

    @property
    def feedable(self):
        return self._feedable

    @feedable.setter
    def feedable(self, val):
        self._feedable = val

    @property
    def type_ref(self):
        return self._type_ref

    @type_ref.setter
    def type_ref(self, val):
        self._type_ref = val

    @property
    def receiver(self):
        return self._receiver

    @receiver.setter
    def receiver(self, val):
        self._receiver = val

    @property
    def obj_builder(self):
        return self._obj_builder

    @obj_builder.setter
    def obj_builder(self, val):
        self._obj_builder = val

    @property
    def skip_comments(self):
        return self._skip_comments

    @skip_comments.setter
    def skip_comments(self, val):
        self._skip_comments = val

    @property
    def queue(self):
        return self._queue

    @queue.setter
    def queue(self, val):
        self._queue = val

    def feed(self, raw):
        """Feed in arbitrary chunks of data"""
        if not self._feedable:
            return False
        self._queue.enqueue(raw)
        for tag, payload in self._queue.dequeue():
            if tag == tags.NOP:
                continue
            self._process(tag, payload)
        return True

    def _process(self, tag, payload):
        # ensure that root data structure is a dict
        if not self._stack and tag not in (tags.DICT,
                                           tags.DICT_EMPTY):
            msg = "The root data structure should be a dict"
            raise errors.Error(msg)
        # interpret tag
        try:
            self._interpret(tag, payload)
        except Exception as e:
            msg = "Error at byte index {}: {}{}"
            str_e = str(e)
            str_e = ": " + str_e if str_e else str_e
            msg = msg.format(self._byte_count, type(e).__name__, str_e)
            raise type(e)(msg)
        self._byte_count += 1 + len(payload)
        # consume block
        while self._consume_block():
            pass

    def _interpret(self, tag, payload):
        if tag in (tags.DICT_EMPTY, tags.LIST_EMPTY,
                   tags.SET_EMPTY):
            self._create_context(tag)
            self._remove_context()
        elif tag in (tags.DICT,
                     tags.LIST,
                     tags.SET,
                     tags.GRID,
                     tags.OBJ,
                     tags.COMPLEX,
                     tags.FLOAT_MISC,
                     tags.FLOAT_1, tags.FLOAT_2, tags.FLOAT_3,
                     tags.FLOAT_1_EXT, tags.FLOAT_2_EXT, tags.FLOAT_3_EXT,
                     tags.DATETIME, tags.DATETIME_EXT,
                     tags.DATE,
                     tags.TIME, tags.TIME_EXT,
                     tags.RADIX_HEX, tags.RADIX_HEX_EXT,
                     tags.RADIX_OCT, tags.RADIX_OCT_EXT,
                     tags.RADIX_BIN, tags.RADIX_BIN_EXT):
            self._create_context(tag)
        # closing container
        elif tag == tags.END:
            self._cleanup_stack()
        # process empty grid
        elif tag == tags.GRID_EMPTY:
            self._update_context(self._unpack_grid(tag))
        # process empty obj
        elif tag == tags.OBJ_EMPTY:
            self._update_context(self._unpack_obj(tag))
        # flags
        elif tag == tags.COMMENT:
            self._update_comment_flag(tag)
        # process payload
        else:
            try:
                data = self._unpack_payload(tag, payload)
            except errors.CommentSkip as e:
                pass
            else:
                self._update_context(data)

    def _create_context(self, tag):
        if tag in (tags.DICT, tags.DICT_EMPTY,
                   tags.LIST, tags.LIST_EMPTY,
                   tags.SET, tags.SET_EMPTY):
            self._create_alpha_context(tag)
        else:
            self._create_beta_context(tag)

    def _create_alpha_context(self, tag):
        if tag in (tags.DICT, tags.DICT_EMPTY):
            datatype = "dict"
            new_container = self._type_ref.dict_type()
        elif tag in (tags.LIST, tags.LIST_EMPTY):
            datatype = "list"
            new_container = self._type_ref.list_type()
        elif tag in (tags.SET, tags.SET_EMPTY):
            datatype = "set"
            new_container = self._type_ref.set_type()
        else:
            msg = "Unknown container tag '{}'".format(tag)
            raise errors.Error(msg)
        if len(self._stack) != 0:
            self._update_context(new_container)
        context = Context(tag, datatype, new_container)
        self._stack.append(context)
        self._data = None

    def _create_beta_context(self, tag):
        if tag == tags.OBJ:
            context = Context(tag, "obj", box.Obj())
        else:
            context = Context(tag, "list", list())
        self._stack.append(context)
        self._block_on = True
        self._block_count += 1

    def _get_context(self):
        try:
            return self._stack[-1]
        except IndexError as e:
            msg = "Inconsistent stack state"
            raise errors.Error(msg)

    def _update_context(self, data):
        context = self._stack[-1]
        is_dict_key_cached = context.is_dict_key_cached
        cached_dict_key = context.cached_dict_key
        # Dict and obj
        if context.datatype == "dict":
            if is_dict_key_cached:
                context.container[cached_dict_key] = data
                context.is_dict_key_cached, context.cached_dict_key = False, None
            else:
                context.is_dict_key_cached, context.cached_dict_key = True, data
        # List
        elif context.datatype == "list":
            context.container.append(data)
        # Set
        elif context.datatype == "set":
            context.container.add(data)
        # Obj
        elif context.datatype == "obj":
            if is_dict_key_cached:
                context.container[cached_dict_key] = data
                context.is_dict_key_cached, context.cached_dict_key = False, None
            else:
                context.is_dict_key_cached, context.cached_dict_key = True, data

    def _remove_context(self):
        if not self._stack:
            return
        context = self._stack[-1]
        data = context.container
        if len(self._stack) == 1:
            self._data = data
            self._stack = list()
            if self._receiver:
                self._receiver(self)
        else:
            del self._stack[-1]

    def _consume_block(self):
        if not self._block_on:
            return False
        context = self._get_context()
        tag, container = context.tag, context.container
        n = len(container)
        if tag == tags.COMPLEX and n == 2:
            data = self._unpack_complex(tag, container)
        elif tag == tags.RADIX_HEX and n == 1:
            data = self._unpack_hex_int(tag, container)
        elif tag == tags.RADIX_HEX_EXT and n == 2:
            data = self._unpack_hex_int(tag, container)
        elif tag == tags.RADIX_OCT and n == 1:
            data = self._unpack_oct_int(tag, container)
        elif tag == tags.RADIX_OCT_EXT and n == 2:
            data = self._unpack_oct_int(tag, container)
        elif tag == tags.RADIX_BIN and n == 1:
            data = self._unpack_bin_int(tag, container)
        elif tag == tags.RADIX_BIN_EXT and n == 2:
            data = self._unpack_bin_int(tag, container)
        elif tag == tags.FLOAT_MISC and n == 1:
            data = self._unpack_float(tag, container)
        elif tag == tags.FLOAT_1 and n == 1:
            data = self._unpack_float(tag, container)
        elif tag == tags.FLOAT_1_EXT and n == 2:
            data = self._unpack_float(tag, container)
        elif tag == tags.FLOAT_2 and n == 2:
            data = self._unpack_float(tag, container)
        elif tag == tags.FLOAT_2_EXT and n == 3:
            data = self._unpack_float(tag, container)
        elif tag == tags.FLOAT_3 and n == 3:
            data = self._unpack_float(tag, container)
        elif tag == tags.FLOAT_3_EXT and n == 4:
            data = self._unpack_float(tag, container)
        elif tag == tags.DATETIME and n == 3:
            data = self._unpack_datetime(tag, container)
        elif tag == tags.DATETIME_EXT and n == 4:
            data = self._unpack_datetime(tag, container)
        elif tag == tags.DATE and n == 2:
            data = self._unpack_date(tag, container)
        elif tag == tags.TIME and n == 2:
            data = self._unpack_time(tag, container)
        elif tag == tags.TIME_EXT and n == 3:
            data = self._unpack_time(tag, container)
        else:
            return False
        # the next line is safe, i.e., if stack were empty,
        # self._get_context() would've raised an exception
        self._remove_block()
        self._update_context(data)
        return True

    def _remove_block(self):
        del self._stack[-1]
        self._block_count -= 1
        if self._block_count == 0:
            self._block_on = False

    def _update_comment_flag(self, tag):
        if tag == tags.COMMENT:
            self._as_comment_str = True

    def _cleanup_stack(self):
        context = self._get_context()
        if context.tag not in (tags.GRID, tags.OBJ):
            self._remove_context()
            return
        if context.tag == tags.GRID:
            data = self._unpack_grid(context.tag, context.container)
        elif context.tag == tags.OBJ:
            data = self._unpack_obj(context.tag, context.container)
        else:
            msg = "END tag is only allowed to end containers, grids, and objs"
            raise errors.Error(msg)
        self._remove_block()
        self._update_context(data)

    def _unpack_payload(self, tag, payload=None):
        # unpack null
        if tag == tags.NULL:
            if self._as_comment_str:
                self._as_comment_str = False
                return self._unpack_comment_id(tags.COMMENT, None)
            return self._unpack_null(tag, payload)
        # boolean
        elif tag in (tags.BOOL_TRUE, tags.BOOL_FALSE):
            return self._unpack_bool(tag, payload)
        # unpack integer
        elif (tags.PINT_8 <= tag <= tags.NINT_HEAVY
              or tags.CONST_0 <= tag <= tags.CONST_99):
            return self._unpack_int(tag, payload)
        # float misc
        #elif tag == tags.FLOAT_MISC:
        #    return self._unpack_float_misc(tag, payload)
        # unpack string
        elif (tags.STR_8 <= tag <= tags.STR_HEAVY
              or tags.CHAR_A <= tag <= tags.CHAR_UP_Z):
            if self._as_comment_str:
                if self._skip_comments:
                    raise errors.CommentSkip
                return self._unpack_comment(tag, payload)
            else:
                return self._unpack_str(tag, payload)
        # unpack bin
        elif tags.BIN_EMPTY <= tag <= tags.BIN_HEAVY:
            return self._unpack_bin(tag, payload)
        # grid's DIV flag
        elif tag == tags.GRID_DIV:
            return GridDiv

    def _unpack_null(self, tag, payload=None):
        return None

    def _unpack_bool(self, tag, payload=None):
        if tag == tags.BOOL_TRUE:
            return self._type_ref.bool_type(True)
        elif tag == tags.BOOL_FALSE:
            return self._type_ref.bool_type(False)

    def _unpack_int(self, tag, payload=None):
        if payload is None:
            msg = "Payload missing"
            raise errors.Error(msg)
        r = unpack_int(tag, payload)
        r = r if self._type_ref.int_type is int else self._type_ref.int_type(r)
        return r

    def _unpack_hex_int(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if tag == tags.RADIX_HEX:
            r = payload_list[0]
        elif tag == tags.RADIX_HEX_EXT:
            leading_zeros, x = payload_list
            r = misc.add_leading_zeros(x, leading_zeros)
        else:
            msg = "Malformed hex_int block"
            raise errors.Error(msg)
        return self._type_ref.hex_int_type(r)

    def _unpack_oct_int(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if tag == tags.RADIX_OCT:
            r = payload_list[0]
        elif tag == tags.RADIX_OCT_EXT:
            leading_zeros, x = payload_list
            r = misc.add_leading_zeros(x, leading_zeros)
        else:
            msg = "Malformed oct_int block"
            raise errors.Error(msg)
        return self._type_ref.oct_int_type(r)

    def _unpack_bin_int(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if tag == tags.RADIX_BIN:
            r = payload_list[0]
        elif tag == tags.RADIX_BIN_EXT:
            leading_zeros, x = payload_list
            r = misc.add_leading_zeros(x, leading_zeros)
        else:
            msg = "Malformed bin_int block"
            raise errors.Error(msg)
        return self._type_ref.bin_int_type(r)

    def _unpack_float(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if tag == tags.FLOAT_MISC:
            return self._unpack_float_misc(tag, payload_list[0])
        elif tag == tags.FLOAT_1:
            significand = payload_list[0]
            return self._type_ref.float_type(significand)
        elif tag == tags.FLOAT_1_EXT:
            significand, exponent = payload_list
            x = "{}E{}".format(significand, exponent)
            return self._type_ref.float_type(x)
        elif tag == tags.FLOAT_2:
            left_significand, right_significand = payload_list
            x = "{}.{}".format(left_significand, right_significand)
            return self._type_ref.float_type(x)
        elif tag == tags.FLOAT_2_EXT:
            left_significand, right_significand, exponent = payload_list
            x = "{}.{}E{}".format(left_significand, right_significand,
                                  exponent)
            return self._type_ref.float_type(x)
        elif tag == tags.FLOAT_3:
            left_significand, leading_zeros, right_significand = payload_list
            leading_zeros = "0" * leading_zeros
            x = "{}.{}{}".format(left_significand, leading_zeros,
                                 right_significand)
            return self._type_ref.float_type(x)
        elif tag == tags.FLOAT_3_EXT:
            left_significand, leading_zeros, right_significand, exponent = payload_list
            leading_zeros = "0" * leading_zeros
            x = "{}.{}{}E{}".format(left_significand, leading_zeros,
                                    right_significand, exponent)
            return self._type_ref.float_type(x)
        else:
            msg = "Malformed float block"
            raise errors.Error(msg)

    def _unpack_float_misc(self, tag, payload=None):
        if payload == "n":
            return self._type_ref.float_type("NaN")
        elif payload == "x":
            return self._type_ref.float_type("+inf")
        elif payload == "y":
            return self._type_ref.float_type("-inf")
        elif payload == "z":
            return self._type_ref.float_type("-0.0")
        else:
            msg = "Malformed float misc datum"
            raise errors.Error(msg)

    def _unpack_complex(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if len(payload_list) == 2:
            real, imag = payload_list
            return self._type_ref.complex_type(real, imag)
        else:
            msg = "Malformed complex number"
            raise errors.Error(msg)

    def _unpack_bin(self, tag, payload=None):
        if payload is None:
            msg = "Payload missing"
            raise errors.Error(msg)
        r = unpack_bin(tag, payload)
        r = r if self._type_ref.bin_type is bytes else self._type_ref.bin_type(r)
        return r

    def _unpack_str(self, tag, payload=None):
        if payload is None:
            msg = "Payload missing"
            raise errors.Error(msg)
        r = unpack_str(tag, payload)
        r = r if self._type_ref.str_type is str else self._type_ref.str_type(r)
        return r

    def _unpack_comment_id(self, tag, payload=None):
        return self._type_ref.comment_id_type()

    def _unpack_comment(self, tag, payload=None):
        if payload is None:
            msg = "Payload missing"
            raise errors.Error(msg)
        self._as_comment_str = False
        r = self._unpack_str(tag, payload)
        return self._type_ref.comment_type(r)

    def _unpack_obj(self, tag, payload_obj=None):
        if tag == tags.OBJ_EMPTY:
            return self._type_ref.obj_type()
        if payload_obj is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if self._obj_builder:
            return self._obj_builder(payload_obj)
        if type(payload_obj) == self._type_ref.obj_type:
            return payload_obj
        return self._type_ref.obj_type(payload_obj)

    def _unpack_grid(self, tag, payload_list=None):
        if tag == tags.GRID_EMPTY:
            return self._type_ref.grid_type()
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        row = list()
        row_size = 0
        grid = list()
        for item in payload_list:
            if item is GridDiv:
                row_size = len(row)
            else:
                if row_size and len(row) == row_size:
                    grid.append(tuple(row))
                    row = list()
                row.append(item)
        if row:
            if row_size and len(row) != row_size:
                msg = "Inconsistent grid"
                raise errors.Error(msg)
            grid.append(tuple(row))
        # check type validity
        for row in grid:
            for item in row:
                dtype = self._type_ref.check(type(item))
                if dtype not in ("int", "float", "complex"):
                    msg = "A grid should be built with int, float, or complex numbers"
                    raise errors.Error(msg)
        return self._type_ref.grid_type(grid)

    def _unpack_datetime(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if len(payload_list) == 3:  # year, nanoseconds, trailing zeros
            year_delta, ns_delta, trailing_zeros = payload_list
            tz_minutes = None
        elif len(payload_list) == 4:  # year, nanoseconds, trailing zeros, timezone
            year_delta, ns_delta, trailing_zeros, tz_minutes = payload_list
        else:
            msg = "Inconsistent datetime"
            raise errors.Error(msg)
        r = misc.construct_datetime(year_delta, ns_delta, trailing_zeros, tz_minutes)
        r = r if self._type_ref.datetime_type is datetime.datetime\
            else self._type_ref.datetime_type(r)
        return r

    def _unpack_date(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if len(payload_list) == 2:  # year_delta, day_delta
            year_delta, day_delta = payload_list
        else:
            msg = "Inconsistent date"
            raise errors.Error(msg)
        r = misc.construct_date(year_delta, day_delta)
        r = r if self._type_ref.date_type is datetime.date \
            else self._type_ref.date_type(r)
        return r

    def _unpack_time(self, tag, payload_list=None):
        if payload_list is None:
            msg = "Payloads missing"
            raise errors.Error(msg)
        if len(payload_list) == 2:  # nanoseconds, trailing zeros
            ns_delta, trailing_zeros = payload_list
            tz_minutes = None
        elif len(payload_list) == 3:  # nanoseconds, trailing zeros, timezone
            ns_delta, trailing_zeros, tz_minutes = payload_list
        else:
            msg = "Inconsistent datetime"
            raise errors.Error(msg)
        r = misc.construct_time(ns_delta, trailing_zeros, tz_minutes)
        r = r if self._type_ref.time_type is datetime.time \
            else self._type_ref.time_type(r)
        return r


class Context:
    def __init__(self, tag, datatype, container):
        self._tag = tag
        self._datatype = datatype
        self._container = container
        self._cached_dict_key = None
        self._is_dict_key_cached = False

    @property
    def tag(self):
        return self._tag

    @property
    def datatype(self):
        return self._datatype

    @property
    def container(self):
        return self._container

    @property
    def cached_dict_key(self):
        return self._cached_dict_key

    @cached_dict_key.setter
    def cached_dict_key(self, val):
        self._cached_dict_key = val

    @property
    def is_dict_key_cached(self):
        return self._is_dict_key_cached

    @is_dict_key_cached.setter
    def is_dict_key_cached(self, val):
        self._is_dict_key_cached = val


class GridDiv:
    pass


def unpack_int(tag, payload):
    if (tags.PINT_8 <= tag <= tags.PINT_HEAVY
            or tags.CONST_0 <= tag <= tags.CONST_99):
        r = unpack_pint(tag, payload)
    # unpack negative integer
    elif tags.NINT_8 <= tag <= tags.NINT_HEAVY:
        r = unpack_nint(tag, payload)
    else:
        msg = "Unknown integer tag byte {}".format(tag)
        raise errors.Error(msg)
    return r


def unpack_str(tag, payload):
    if tag == tags.STR_EMPTY:
        return str()
    elif tags.CHAR_A <= tag <= tags.CHAR_UP_Z:
        for k, v in ALPHABET.items():
            if v == tag:
                return k
    else:
        return payload.decode("utf-8")


def unpack_bin(tag, payload):
    if tag == tags.BIN_EMPTY:
        return b''
    else:
        return bytes(payload)


def unpack_pint(tag, payload):
    if tags.CONST_0 <= tag <= tags.CONST_99:
        return 99 - (tags.CONST_99[0] - tag[0])
    return int.from_bytes(payload, "little", signed=False)


def unpack_nint(tag, payload):
    return -int.from_bytes(payload, "little", signed=False)



import unittest
import datetime
from textwrap import dedent
from paradict import box, errors, deserializer
from paradict.serializer.encoder import Encoder


class TestEmptyData(unittest.TestCase):

    def test(self):
        d = ""
        r = decode_data(dedent(d))
        expected = dict()
        self.assertEqual(expected, r)


class TestDictInConfigMode(unittest.TestCase):

    def test_valid_dict(self):
        d = """\
        a_valid_key_42 = (dict)
            another_key = 42
            null = null
            true = true
        """
        r = decode_data(dedent(d))
        expected = {"a_valid_key_42": {"another_key": 42,
                                       "null": None,
                                       "true": True}}
        self.assertEqual(expected, r)

    def test_invalid_dict(self):
        d1 = """4 2 = null"""
        d2 = """4.2 = null"""
        d3 = """1+2i = null"""
        for i, d in enumerate((d1, d2, d3)):
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    decode_data(dedent(d))

    def test_malformed_dict_key(self):
        d1 = """key_42= = null"""
        d2 = """1 key_42 = null"""
        for i, d in enumerate((d1, d2)):
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    r = decode_data(dedent(d))
                    expected = {0: dict()}
                    self.assertEqual(expected, r)


class TestDict(unittest.TestCase):

    def test_empty_dict(self):
        d = """\
        0: (dict)
        """
        r = decode_data(dedent(d))
        expected = {0: dict()}
        self.assertEqual(expected, r)

    def test_dict_with_valid_keys(self):
        d = """\
        0: null
        "1": null
        "": null
        "key": null
        3.14: null
        1+2i: null
        "\\n": null
        "\\\\n": null
        """
        r = decode_data(dedent(d))
        expected = {0: None,
                    "1": None,
                    "": None,
                    "key": None,
                    3.14: None,
                    complex(1, 2): None,
                    "\n": None,
                    "\\n": None}
        self.assertEqual(expected, r)

    def test_dict_with_valid_values(self):
        d = """\
        0: "hello world"
        1: (text)
            multiline
            string
            ---
        2: true
        3: false
        4: 1+2i
        5: 42
        6: 4.2
        7: 2020-12-31
        """
        r = decode_data(dedent(d))
        expected = {0: "hello world",
                    1: "multiline\nstring",
                    2: True,
                    3: False,
                    4: complex(1, 2),
                    5: 42,
                    6: 4.2,
                    7: datetime.date(2020, 12, 31)}
        self.assertEqual(expected, r)

    def test_dict_with_invalid_keys(self):
        d1 = """key_42: null"""
        d2 = """2020-12-31: null"""
        d3 = """10:12:59 : null"""
        d4 = """null : null"""
        d5 = """true : null"""
        for i, d in enumerate((d1, d2, d3, d4, d5)):
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    decode_data(dedent(d))

    def test_nested_dict(self):
        d = """\
        0: "hello world"
        "1": (text)
            multiline
            string
            ---
        "": true
        "key": false
        3.14: 1+2i
        1+2i: 42
        "\\n": 4.2
        "\\\\n": 2020-12-31
        "nested": (dict)
            0: "hello world"
            "1": (text)
                multiline
                string
                ---
            "": true
            "key": false
            3.14: 1+2i
            1+2i: 42
            "\\n": 4.2
            "\\\\n": 2020-12-31
        """
        r = decode_data(dedent(d))
        expected = {0: "hello world",
                    "1": "multiline\nstring",
                    "": True,
                    "key": False,
                    3.14: complex(1, 2),
                    complex(1, 2): 42,
                    "\n": 4.2,
                    "\\n": datetime.date(2020, 12, 31),
                    "nested": {0: "hello world",
                               "1": "multiline\nstring",
                               "": True,
                               "key": False,
                               3.14: complex(1, 2),
                               complex(1, 2): 42,
                               "\n": 4.2,
                               "\\n": datetime.date(2020, 12, 31)}}
        self.assertEqual(expected, r)


class TestList(unittest.TestCase):

    def test_empty_list(self):
        d = """\
        0: (list)
        """
        r = decode_data(dedent(d))
        expected = {0: list()}
        self.assertEqual(expected, r)

    def test_list_with_valid_data(self):
        d = """\
        0: (list)
            0
            4.2
            "1"
            true
            false
            null
            1+2i
            "hello world"
            2020-12-31
            (text)
                multiline
                text
                ---
        """
        r = decode_data(dedent(d))
        expected = {0: [0,
                        4.2,
                        "1",
                        True,
                        False,
                        None,
                        complex(1, 2),
                        "hello world",
                        datetime.date(2020, 12, 31),
                        "multiline\ntext"]}
        self.assertEqual(expected, r)

    def test_nested_list(self):
        d = """\
        0: (list)
            0
            4.2
            "1"
            true
            false
            null
            1+2i
            "hello world"
            2020-12-31
            (text)
                multiline
                text
                ---
            (list)
                0
                4.2
                "1"
                true
                false
                null
                1+2i
                "hello world"
                2020-12-31
                (text)
                    multiline
                    text
                    ---
        """
        r = decode_data(dedent(d))
        expected = {0: [0,
                        4.2,
                        "1",
                        True,
                        False,
                        None,
                        complex(1, 2),
                        "hello world",
                        datetime.date(2020, 12, 31),
                        "multiline\ntext",
                        [0,
                         4.2,
                         "1",
                         True,
                         False,
                         None,
                         complex(1, 2),
                         "hello world",
                         datetime.date(2020, 12, 31),
                         "multiline\ntext"]]}
        self.assertEqual(expected, r)


class TestSet(unittest.TestCase):

    def test_empty_set(self):
        d = """\
        0: (set)
        """
        r = decode_data(dedent(d))
        expected = {0: set()}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], set)

    def test_set_deletable(self):
        d = """\
        0: (set)
            (text)
                multiline
                text
                ---
        """
        r = decode_data(dedent(d))
        expected = {0: {"multiline\ntext"}}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], set)

    def test_set_with_valid_data(self):
        d = """\
        0: (set)
            0
            4.2
            "1"
            1+2i
            "hello world"
            2020-12-31
            (text)
                multiline
                text
                ---
        """
        r = decode_data(dedent(d))
        expected = {0: {0,
                        4.2,
                        "1",
                        complex(1, 2),
                        "hello world",
                        datetime.date(2020, 12, 31),
                        "multiline\ntext"}}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], set)


class TestObj(unittest.TestCase):

    def test_empty_obj(self):
        d = """\
        0: (obj)
        """
        r = decode_data(dedent(d))
        expected = {0: box.Obj()}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], box.Obj)

    def test_obj_with_valid_keys(self):
        d = """\
        0: (obj)
            0: null
            "1": null
            "": null
            "key": null
            3.14: null
            1+2i: null
            "\\n": null
            "\\\\n": null
        """
        r = decode_data(dedent(d))
        expected = {0: box.Obj({0: None,
                                "1": None,
                                "": None,
                                "key": None,
                                3.14: None,
                                complex(1, 2): None,
                                "\n": None,
                                "\\n": None})}
        self.assertEqual(expected, r)

    def test_obj_with_valid_values(self):
        d = """\
        0: (obj)
            0: "hello world"
            1: (text)
                multiline
                string
                ---
            2: true
            3: false
            4: 1+2i
            5: 42
            6: 4.2
            7: 2020-12-31
        """
        r = decode_data(dedent(d))
        expected = {0: box.Obj({0:"hello world",
                                1: "multiline\nstring",
                                2: True,
                                3: False,
                                4: complex(1, 2),
                                5: 42,
                                6: 4.2,
                                7: datetime.date(2020, 12, 31)})}
        self.assertEqual(expected, r)

    def test_obj_with_invalid_keys(self):
        d1 = """\
        0: (obj)
            key_42: null
        """
        d2 = """\
        0: (obj)
            2020-12-31: null
        """
        d3 = """\
        0: (obj)
            10:12:59 : null
        """
        d4 = """\
        0: (obj)
            null : null
        """
        d5 = """\
        0: (obj)
            true : null
        """
        for i, d in enumerate((d1, d2, d3, d4, d5)):
            with self.subTest("Test {}".format(i + 1)):
                with self.assertRaises(errors.Error):
                    decode_data(dedent(d))

    def test_nested_obj(self):
        d = """\
        0: "hello world"
        "1": (text)
            multiline
            string
            ---
        "": true
        "key": false
        3.14: 1+2i
        1+2i: 42
        "\\n": 4.2
        "\\\\n": 2020-12-31
        "nested": (obj)
            0: (obj)
                0: "hello world"
                "1": (text)
                    multiline
                    string
                    ---
                "": true
                "key": false
                3.14: 1+2i
                1+2i: 42
                "\\n": 4.2
                "\\\\n": 2020-12-31
        """
        r = decode_data(dedent(d))
        expected = {0: "hello world",
                    "1": "multiline\nstring",
                    "": True,
                    "key": False,
                    3.14: complex(1, 2),
                    complex(1, 2): 42,
                    "\n": 4.2,
                    "\\n": datetime.date(2020, 12, 31),
                    "nested": box.Obj({0: box.Obj({0: "hello world",
                               "1": "multiline\nstring",
                               "": True,
                               "key": False,
                               3.14: complex(1, 2),
                               complex(1, 2): 42,
                               "\n": 4.2,
                               "\\n": datetime.date(2020, 12, 31)})})}
        self.assertEqual(expected, r)


class TestObjBuilder(unittest.TestCase):

    def test_simple_obj(self):
        d = """\
        0: (obj)
            'type': 'complex'
            'real': -3.14
            'imag': 4.2
        """
        r = decode_data(dedent(d),
                        obj_builder=lambda obj:
                          complex(obj["real"], obj["imag"]))
        expected = {0: complex(-3.14, 4.2)}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], complex)

    def test_nested_obj(self):
        d = """\
        0: (obj)
            'type': 'complex'
            'real': (obj)
                'type': 'number'
                'x': -3.14
            'imag': (obj)
                'type': 'number'
                'x': 4.2
        """
        r = decode_data(dedent(d), obj_builder=my_obj_builder)
        expected = {0: complex(-3.14, 4.2)}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], complex)


class TestGrid(unittest.TestCase):

    def test_empty_grid(self):
        d = """\
        0: (grid)
        """
        r = decode_data(dedent(d))
        expected = {0: box.Grid()}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_integer(self):  #
        d = """\
        0: (grid)
            0 1 2
            3 4 5_000_000_000
            6 7 -8
        """
        r = decode_data(dedent(d))
        my_grid = [(0, 1, 2),
                   (3, 4, 5_000_000_000),
                   (6, 7, -8)]
        expected = {0: my_grid}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_float(self):
        d = """\
        0: (grid)
            0.0 1.1 -2.2_000_000_000
        """
        r = decode_data(dedent(d))
        my_grid = [(0.0, 1.1, -2.2_000_000_000)]
        expected = {0: my_grid}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_complex(self):
        d = """\
        0: (grid)
            1+2i 3+4.5_000_000_000i
        """
        r = decode_data(dedent(d))
        my_grid = [(1+2j, 3+4.5_000_000_000j)]
        expected = {0: my_grid}
        self.assertEqual(expected, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_illegal_data(self):
        d = """\
        0: (grid)
            "a" "b"
        """
        with self.assertRaises(errors.Error):
            decode_data(dedent(d))


class TestComment(unittest.TestCase):

    def test_empty_comment(self):
        d = """\
        #
        #
        0: (dict)
            #
            #
            0: (list)
                #
                #
                (obj)
                    #
                    #
                    0: (set)
                        #"""
        r = decode_encode(dedent(d), skip_comments=False)
        self.assertEqual(dedent(d), r)

    def test_regular_comment(self):
        d = """\
        # comment 1
        0: (dict)
            # comment 2
            0: (list)
                # comment 3
                (obj)
                    # comment 4
                    0: (set)
                        # comment 5"""
        r = decode_encode(dedent(d), skip_comments=False)
        self.assertEqual(dedent(d), r)


class TestDatetime(unittest.TestCase):

    def test_naive_datetime_lower_endpoint(self):
        d = """0: 0001-01-01T00:00:00"""
        r = decode_data(d)
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_naive_datetime_upper_endpoint(self):
        d = "0: 9999-12-31T23:59:59.999999"
        r = decode_data(d)
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_aware_datetime_lower_endpoint(self):
        d = "0: 0001-01-01T00:00:00Z"
        r = decode_data(d)
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_aware_datetime_upper_endpoint(self):
        d = "0: 9999-12-31T23:59:59.999999+23:59"
        r = decode_data(d)
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_aware_datetime_with_negative_negative_utc_timezone(self):
        d = "0: 9999-12-31T23:59:59.999999-01:30"
        r = decode_data(d)
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=-1, minutes=-30))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        expected = {0: val}
        self.assertEqual(expected, r)


class TestDate(unittest.TestCase):

    def test_date_lower_endpoint(self):
        d = """0: 0001-01-01"""
        r = decode_data(dedent(d))
        year, month, day = 1, 1, 1
        val = datetime.date(year=year, month=month, day=day)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_date_upper_endpoint(self):
        d = """0: 9999-12-31"""
        r = decode_data(dedent(d))
        year, month, day = 9999, 12, 31
        val = datetime.date(year=year, month=month, day=day)
        expected = {0: val}
        self.assertEqual(expected, r)


class TestTime(unittest.TestCase):

    def test_naive_time_lower_endpoint(self):
        d = """0: 00:00:00"""
        r = decode_data(d)
        hour, minute = 0, 0
        second, microsecond = 0, 0
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_naive_time_upper_endpoint(self):
        d = "0: 23:59:59.999999"
        r = decode_data(d)
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_aware_time_lower_endpoint(self):
        d = "0: 00:00:00Z"
        r = decode_data(d)
        hour, minute = 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=0, minutes=0))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_aware_time_upper_endpoint(self):
        d = "0: 23:59:59.999999+23:59"
        r = decode_data(d)
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        expected = {0: val}
        self.assertEqual(expected, r)

    def test_aware_time_with_negative_negative_utc_timezone(self):
        d = "0: 23:59:59.999999-01:30"
        r = decode_data(d)
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=-1, minutes=-30))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        expected = {0: val}
        self.assertEqual(expected, r)


class TestNull(unittest.TestCase):

    def test(self):
        d = """\
        0: null
        """
        r = decode_data(dedent(d))
        expected = {0: None}
        self.assertEqual(expected, r)


class TestBool(unittest.TestCase):

    def test(self):
        d = """\
        0: true
        1: false
        """
        r = decode_data(dedent(d))
        expected = {0: True, 1: False}
        self.assertEqual(expected, r)


class TestBin(unittest.TestCase):

    def test(self):
        d = """\
        0: (bin)
            48 65 6C 6C 6F 20 57 6F 72 6C 64 20 21 20 48 65
            6C 6C 6F 20 57 6F 72 6C 64 20 21
        1: (bin)"""
        bin_hello = b'Hello World ! Hello World !'
        r = decode_data(dedent(d))
        expected = {0: bin_hello, 1: b''}
        self.assertEqual(expected, r)


class TestInt(unittest.TestCase):

    def test_int(self):
        d = """\
        0: (list)
            0
            1
            2_000_000
        """
        r = decode_data(dedent(d))
        expected = {0: [0, 1, 2_000_000]}
        self.assertEqual(expected, r)

    def test_hex_oct_bin_int(self):
        d = """\
        0: (list)
            0xffff_ffff
            0o37_777_777_777
            0b1111_1111_1111_1111_1111_1111_1111_1111
        """
        r = decode_data(dedent(d))
        x = 2**32-1
        expected = {0: [box.HexInt(x),
                        box.OctInt(x),
                        box.BinInt(x)]}
        self.assertEqual(expected, r)


class TestFloat(unittest.TestCase):

    def test(self):
        d = """\
        0: (list)
            0.0
            1.0
            2_000_000.000_001
            3.14E10
            3.14E-10
            3.000001
            0.0004
            0.0009E10
            -0.0009E-10
        """
        r = decode_data(dedent(d))
        expected = {0: [float("0.0"),
                        float("1.0"),
                        float("2_000_000.000_001"),
                        float("3.14E10"),
                        float("3.14E-10"),
                        float("3.000001"),
                        float("0.0004"),
                        float("0.0009E10"),
                        float("-0.0009E-10")]}
        self.assertEqual(expected, r)


class TestComplex(unittest.TestCase):

    def test(self):
        d = """\
        0: (list)
            1+2i
            -3-4i
            2_000_000i
            3.14+1i
        """
        r = decode_data(dedent(d))
        expected = {0: [complex(1, 2),
                        complex(-3, -4),
                        complex(0, 2000000),
                        complex(3.14, 1)]}
        self.assertEqual(expected, r)

    def test_hex_oct_bin_int(self):
        d = """\
        0: (list)
            0xffff_ffff
            0o37_777_777_777
            0b1111_1111_1111_1111_1111_1111_1111_1111
        """
        r = decode_data(dedent(d))
        x = 2**32-1
        expected = {0: [box.HexInt(x),
                        box.OctInt(x),
                        box.BinInt(x)]}
        self.assertEqual(expected, r)


class TestString(unittest.TestCase):

    def test_ordinary_str(self):
        d = """\
        0: (list)
            "hello˫ world\\u02eb"
            "hello \\n world"
            "hello \\\\ world\\\\u02eb"
        """
        r = decode_data(dedent(d))
        expected = {0: ["hello˫ world\u02eb", "hello \n world",
                        "hello \\ world\\u02eb"]}
        self.assertEqual(expected, r)

    def test_raw_str(self):
        d = """\
        0: (list)
            'hello˫ world\\u02eb'
            'hello \\n world'
            'hello \\\\ world\\\\u02eb'
        """
        r = decode_data(dedent(d))
        expected = {0: ["hello˫ world\\u02eb", "hello \\n world",
                        "hello \\\\ world\\\\u02eb"]}
        self.assertEqual(expected, r)


class TestMultilineString(unittest.TestCase):

    def test_ordinary_str(self):
        d = """\
        0: (text)
            Hello˫ world
               ***
                ***
            This\\u02eb is a \\n multiline
            string\\\\
            .
            
            
            ---
        """
        r = decode_data(dedent(d))
        expected = {0: "Hello˫ world\n   ***\n    ***\n"
                       "This\u02eb is a \n multiline\nstring\\\n.\n\n"}
        self.assertEqual(expected, r)

    def test_raw_str(self):
        d = """\
        0: (raw)
            Hello˫ world
               ***
                ***
            This\\u02eb is a \\n multiline
            string\\\\
            .


            ---
        """
        r = decode_data(dedent(d))
        expected = {0: "Hello˫ world\n   ***\n    ***\n"
                       "This\\u02eb is a \\n multiline\nstring\\\\\n.\n\n"}
        self.assertEqual(expected, r)


class TestRaw(unittest.TestCase):

    def test_raw_string(self):
        d = """\
        0: (list)
            'hello˫ world\\u02eb'
            'hello \\n world'
            'hello \\ world\\\\u02eb'
        """
        r = decode_data(dedent(d))
        expected = {0: ["hello˫ world\\u02eb", "hello \\n world",
                        "hello \\ world\\\\u02eb"]}
        self.assertEqual(expected, r)
        for i, item in enumerate(r[0]):
            with self.subTest("subtest index: {}".format(i)):
                self.assertEqual(str, type(item))

    def test_multiline_raw_string(self):
        d = """\
        0: (raw)
            Hello˫ world
               ***
                ***
            This\\u02eb is a \\n multiline
            string\\
            .
            
            
            ---
        """
        r = decode_data(dedent(d))
        expected = {0: "Hello˫ world\n   ***\n    ***\n"
                       "This\\u02eb is a \\n multiline\nstring\\\n.\n\n"}
        self.assertEqual(expected, r)


def decode_data(s, **kwargs):
    return deserializer.decode(s, **kwargs)


def decode_encode(s, **kwargs):
    data = decode_data(s, **kwargs)
    buffer = list()
    encoder = Encoder(**kwargs)
    for line in encoder.encode(data):
        buffer.append(line)
    return "\n".join(buffer)


def my_obj_builder(obj):
    datatype = obj["type"]
    if datatype == "number":
        return obj["x"]
    elif datatype == "complex":
        return complex(obj["real"], obj["imag"])


if __name__ == '__main__':
    unittest.main()



"""Boxes to hold comments, grids, hexadecimal integers, et cetera"""
from paradict import misc
from ustrid import ustrid


class Grid(list):
    """
    Box to hold a grid. A grid is made of numbers and should
    be consistent (rows should be of same size)

    Example:
        # a grid with 2 rows and 3 columns
        my_grid = Grid([(0, 1, 0),
                        (1, 0, 1)])
    """
    pass


class Obj(dict):
    """
    Box to hold an Extension Object. Such objects behave like a dictionary.
    An object builder is passed as argument to the right functions, thus,
    the object is consumed/used to build a new valid data value
    """
    pass


class HexInt(int):
    """Box to hold hexadecimal integer"""

    def __new__(cls, x):
        width = 0
        if isinstance(x, str):
            parts = misc.split_int(x)
            if parts.prefix == "0x":
                width = len(parts.leading_zeros) + len(parts.val)
        instance = super().__new__(cls, str(x), base=misc.get_int_base(x))
        x = hex(instance)
        x = misc.left_pad_int(x, width) if width else x
        instance.__x = misc.tidy_up_int(x, width=4)
        return instance

    def __str__(self):
        return self.__x


class OctInt(int):
    """Box to hold octal integer"""

    def __new__(cls, x):
        width = 0
        if isinstance(x, str):
            parts = misc.split_int(x)
            if parts.prefix == "0o":
                width = len(parts.leading_zeros) + len(parts.val)
        instance = super().__new__(cls, str(x), base=misc.get_int_base(x))
        x = oct(instance)
        x = misc.left_pad_int(x, width) if width else x
        instance.__x = misc.tidy_up_int(x, width=3)
        return instance

    def __str__(self):
        return self.__x


class BinInt(int):
    """Box to hold binary integer"""

    def __new__(cls, x):
        width = 0
        if isinstance(x, str):
            parts = misc.split_int(x)
            if parts.prefix == "0b":
                width = len(parts.leading_zeros) + len(parts.val)
        instance = super().__new__(cls, str(x), base=misc.get_int_base(x))
        x = bin(instance)
        x = misc.left_pad_int(x, width) if width else x
        instance.__x = misc.tidy_up_int(x, width=4)
        return instance

    def __str__(self):
        return self.__x


class CommentID(str):
    """
    Box to hold a unique comment id.
    Instantiating this class will generate a new unique string.
    Under the hood, the 'ustrid' library is used.
        CommentID() != CommentID()
    """
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls, ustrid())


class Comment(str):
    """Box to hold a comment string"""
    pass



import unittest
import pathlib
import tempfile
from paradict import load, dump


class TestLoadFunc(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path = file.name

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test(self):
        data = {0: "hello world", "pi": 3.14}
        dump(data, self._path)
        r = load(self._path)
        expected = data
        self.assertEqual(expected, r)


class TestDecodeFunc(unittest.TestCase):
    """The 'decode' function is already heavily
     used in the 'decoder' test module"""
    def test(self):
        self.assertTrue(True)


class TestPackFunc(unittest.TestCase):
    """The 'unpack' function is already heavily
     used in the 'unpacker' test module"""
    def test(self):
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



import unittest


class TestImports(unittest.TestCase):

    def test_import_funcs(self):
        try:
            # import functions
            from paradict import encode
            from paradict import decode
            from paradict import pack
            from paradict import unpack
            from paradict import load
            from paradict import dump
            from paradict import forge_bin
            from paradict import stringify_bin
            from paradict import validate
            from paradict import split_kv
        except ImportError:
            self.assertTrue(False)

    def test_import_classes(self):
        try:
            # import classes
            from paradict import Encoder
            from paradict import Decoder
            from paradict import Packer
            from paradict import Unpacker
            from paradict import TypeRef
            from paradict import Validator
        except ImportError:
            self.assertTrue(False)


if __name__ == '__main__':
    unittest.main()


"""High-level functions to deserialize Paradict binary/text data into a Python dict"""
import pathlib
from paradict.deserializer.decoder import Decoder
from paradict.deserializer.unpacker import Unpacker


def decode(text, type_ref=None, receiver=None, obj_builder=None,
           skip_comments=False):
    """
    Convert some textual Paradict data into a Python dictionary

    [parameters]
    - text: string to convert into a Python dict
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.box.Obj container and
    returns a fresh new Python object
    - skip_comments: boolean to tell whether comments should be ignored or not

    [return]
    Return the newly built Python object
    """
    decoder = Decoder(type_ref=type_ref, receiver=receiver,
                      obj_builder=obj_builder,
                      skip_comments=skip_comments)
    decoder.feed(text)
    if decoder.queue.buffer:
        decoder.feed("\n")
    decoder.feed("===\n")
    return decoder.data


def unpack(raw, type_ref=None, receiver=None, obj_builder=None,
           skip_comments=False):
    """
    Convert some binary Paradict data into a Python dictionary

    [parameters]
    - raw: raw data previously packed with Paradict
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.box.Obj container and
    returns a fresh new Python object
    - skip_comments: boolean to tell whether comments should be ignored or not

    [return]
    Return the newly built Python object
    """
    unpacker = Unpacker(type_ref=type_ref,
                        receiver=receiver,
                        obj_builder=obj_builder,
                        skip_comments=skip_comments)
    unpacker.feed(raw)
    return unpacker.data


def load(path, type_ref=None, obj_builder=None, skip_comments=False):
    """
    Open a binary Paradict file then unpack its contents into Python dict

    [parameters]
    - path: a path string, or a pathlib.Path instance
    - type_ref: optional TypeRef object
    - receiver: callback function that will be called at the end of conversion.
    This callback function accepts the Decoder instance as argument
    - obj_builder: function that accepts a paradict.box.Obj container and
    returns a fresh new Python object
    - skip_comments: boolean to tell whether comments should be ignored or not

    [return]
    Return the newly built Python object
    """
    path = str(pathlib.Path(path).resolve())
    with open(path, "rb") as file:
        r = file.read()
    return unpack(r, obj_builder=obj_builder, type_ref=type_ref,
                  skip_comments=skip_comments)



import unittest
import pathlib
import tempfile
from paradict import dump, pack


class TestDumpFunc(unittest.TestCase):

    def setUp(self):
        file = tempfile.NamedTemporaryFile(delete=False)
        file.close()
        self._path = file.name

    def tearDown(self):
        pathlib.Path(self._path).unlink()

    def test(self):
        data = {0: "hello world", "pi": 3.14}
        dump(data, self._path)
        with open(self._path, "rb") as file:
            r = file.read()
        expected = pack(data)
        self.assertEqual(expected, r)


class TestEncodeFunc(unittest.TestCase):
    """The 'encode' function is already heavily
     used in the 'encoder' test module"""
    def test(self):
        self.assertTrue(True)


class TestPackFunc(unittest.TestCase):
    """The 'pack' function is already heavily
     used in the 'packer' test module"""
    def test(self):
        self.assertTrue(True)


if __name__ == '__main__':
    unittest.main()



import unittest
from paradict.queue.bin_queue import BinQueue
from paradict import misc, tags


class TestQueueWithINT(unittest.TestCase):

    def test_complete_pint_datum(self):
        data = misc.forge_bin(tags.PINT_BIG, 1, b'\x00' * 2)
        queue = BinQueue()
        r = put_and_get(queue, data)
        expected = misc.forge_bin(tags.PINT_BIG, b'\x00' * 2)
        self.assertEqual(expected, r)
        self.assertEqual(0, len(queue.buffer))

    def test_complete_nint_datum(self):
        data = misc.forge_bin(tags.NINT_BIG, 1, b'\x00' * 2)
        queue = BinQueue()
        r = put_and_get(queue, data)
        expected = misc.forge_bin(tags.NINT_BIG, b'\x00' * 2)
        self.assertEqual(expected, r)
        self.assertEqual(0, len(queue.buffer))

    def test_complete_pintx_datum(self):
        i = 0
        for tag in [tags.PINT_8, tags.PINT_16, tags.PINT_24,
                    tags.PINT_32, tags.PINT_40, tags.PINT_48,
                    tags.PINT_56, tags.PINT_64]:
            i += 8
            with self.subTest("PINT_{} tag".format(i)):
                n = i//8
                data = misc.forge_bin(tag, b'\x00'*n)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertEqual(data, r)
                self.assertEqual(0, len(queue.buffer))

    def test_complete_nintx_datum(self):
        i = 0
        for tag in [tags.NINT_8, tags.NINT_16, tags.NINT_24,
                    tags.NINT_32, tags.NINT_40, tags.NINT_48,
                    tags.NINT_56, tags.NINT_64]:
            i += 8
            with self.subTest("NINT_{} tag".format(i)):
                n = i//8
                data = misc.forge_bin(tag, b'\x00'*n)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertEqual(data, r)
                self.assertEqual(0, len(queue.buffer))

    def test_incomplete_pint_datum(self):
        data = misc.forge_bin(tags.PINT_BIG, 1, b'\x00' * 1)
        queue = BinQueue()
        r = put_and_get(queue, data)
        self.assertIsNone(r)
        self.assertNotEqual(0, len(queue.buffer))

    def test_incomplete_nint_datum(self):
        data = misc.forge_bin(tags.NINT_BIG, 1, b'\x00' * 1)
        queue = BinQueue()
        r = put_and_get(queue, data)
        self.assertIsNone(r)
        self.assertNotEqual(0, len(queue.buffer))

    def test_incomplete_pintx_datum(self):
        i = 0
        for tag in [tags.PINT_8, tags.PINT_16, tags.PINT_24,
                    tags.PINT_32, tags.PINT_40, tags.PINT_48,
                    tags.PINT_56, tags.PINT_64]:
            i += 8
            with self.subTest("PINT_{} tag".format(i)):
                n = (i//8) - 1
                data = misc.forge_bin(tag, b'\x00'*n)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertIsNone(r)
                self.assertNotEqual(0, len(queue.buffer))

    def test_incomplete_nintx_datum(self):
        i = 0
        for tag in [tags.NINT_8, tags.NINT_16, tags.NINT_24,
                    tags.NINT_32, tags.NINT_40, tags.NINT_48,
                    tags.NINT_56, tags.NINT_64]:
            i += 8
            with self.subTest("NINT_{} tag".format(i)):
                n = (i//8) - 1
                data = misc.forge_bin(tag, b'\x00'*n)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertIsNone(r)
                self.assertNotEqual(0, len(queue.buffer))


class TestQueueWithSTR(unittest.TestCase):

    def test_complete_str_datum(self):
        i = 0
        for name, tag in {"STR_SHORT": tags.STR_SHORT,
                          "STR_MEDIUM": tags.STR_MEDIUM,
                          "STR_LONG": tags.STR_LONG}.items():
            i += 8
            with self.subTest("{} tag".format(name)):
                n = i // 8
                size = 2**(8*n)
                data = misc.forge_bin(tag, size-1,  b'\x00' * size)
                queue = BinQueue()
                r = put_and_get(queue, data)
                expected = misc.forge_bin(tag, b'\x00' * size)
                self.assertEqual(expected, r)
                self.assertEqual(0, len(queue.buffer))

    def test_complete_strx_datum(self):
        i = 0
        for tag in [tags.STR_8, tags.STR_16, tags.STR_24,
                    tags.STR_32, tags.STR_40, tags.STR_48,
                    tags.STR_56, tags.STR_64, tags.STR_72,
                    tags.STR_80, tags.STR_88, tags.STR_96,
                    tags.STR_104, tags.STR_112, tags.STR_120,
                    tags.STR_128, tags.STR_136, tags.STR_144,
                    tags.STR_152, tags.STR_160, tags.STR_168,
                    tags.STR_176, tags.STR_184, tags.STR_192,
                    tags.STR_200, tags.STR_208]:
            i += 8
            with self.subTest("STR_{} tag".format(i)):
                n = i // 8
                data = misc.forge_bin(tag, b'\x00' * n)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertEqual(data, r)
                self.assertEqual(0, len(queue.buffer))

    def test_incomplete_str_datum(self):
        i = 0
        for name, tag in {"STR_SHORT": tags.STR_SHORT,
                          "STR_MEDIUM": tags.STR_MEDIUM,
                          "STR_LONG": tags.STR_LONG}.items():
            i += 8
            with self.subTest("{} tag".format(name)):
                n = (i // 8) - 1
                size = 2**(8*n)
                data = misc.forge_bin(tag, size-1,  b'\x00' * size)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertIsNone(r)
                self.assertNotEqual(0, len(queue.buffer))

    def test_incomplete_strx_datum(self):
        i = 0
        for tag in [tags.STR_8, tags.STR_16, tags.STR_24,
                    tags.STR_32, tags.STR_40, tags.STR_48,
                    tags.STR_56, tags.STR_64, tags.STR_72,
                    tags.STR_80, tags.STR_88, tags.STR_96,
                    tags.STR_104, tags.STR_112, tags.STR_120,
                    tags.STR_128, tags.STR_136, tags.STR_144,
                    tags.STR_152, tags.STR_160, tags.STR_168,
                    tags.STR_176, tags.STR_184, tags.STR_192,
                    tags.STR_200, tags.STR_208]:
            i += 8
            with self.subTest("STR_{} tag".format(i)):
                n = (i // 8) - 1
                data = misc.forge_bin(tag, b'\x00' * n)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertIsNone(r)
                self.assertNotEqual(0, len(queue.buffer))


class TestQueueWithBIN(unittest.TestCase):

    def test_complete_bin_datum(self):
        i = 0
        for name, tag in {"BIN_SHORT": tags.BIN_SHORT,
                          "BIN_MEDIUM": tags.BIN_MEDIUM,
                          "BIN_LONG": tags.BIN_LONG}.items():
            i += 8
            with self.subTest("{} tag".format(name)):
                n = i // 8
                size = 2**(8*n)
                data = misc.forge_bin(tag, size-1,  b'\x00' * size)
                queue = BinQueue()
                r = put_and_get(queue, data)
                expected = misc.forge_bin(tag, b'\x00' * size)
                self.assertEqual(expected, r)
                self.assertEqual(0, len(queue.buffer))

    def test_incomplete_bin_datum(self):
        i = 0
        for name, tag in {"BIN_SHORT": tags.BIN_SHORT,
                          "BIN_MEDIUM": tags.BIN_MEDIUM,
                          "BIN_LONG": tags.BIN_LONG}.items():
            i += 8
            with self.subTest("{} tag".format(name)):
                n = (i // 8) - 1
                size = 2**(8*n)
                data = misc.forge_bin(tag, size-1,  b'\x00' * size)
                queue = BinQueue()
                r = put_and_get(queue, data)
                self.assertIsNone(r)
                self.assertNotEqual(0, len(queue.buffer))


def put_and_get(queue, raw):
    queue.enqueue(raw)
    buffer = bytearray()
    for tag, payload in queue.dequeue():
        buffer.extend(tag)
        buffer.extend(payload)
    return None if not buffer else buffer


if __name__ == "__main__":
    unittest.main()



"""TypeRef for type specs and customization"""
import datetime
from decimal import Decimal
from collections import OrderedDict
from paradict import box, errors


class TypeRef:
    """This class represents a mechanism for customizing
    Python types allowed for (de)serializing data with Paradict classes and functions.
    For example, one might want to only use Python OrderedDict instead of the regular
    dict. In this case, just create a TypeRef instance, and make sure that you
    set the dict_type attribute via the construction or a property.

    ```python
    type_ref = TypeRef(dict_type=OrderedDict)
    ```

    Still with this class, one could 'adapt' some exotic datatype so it will
    conform with Python datatypes allowed in Paradict (de)serialization.
    To do so, set the adapters attribute like this:

    ```python
    adapters = {MyExoticType1: adapterFunction1, MyExoticType2: adapterFunction2}
    type_ref = TypeRef(adapters=adapters)
    ```
    """
    def __init__(self, adapters=None,

                 dict_type=None, list_type=None, set_type=None,
                 obj_type=None,

                 dict_types=None, list_types=None, set_types=None,
                 obj_types=None,

                 bin_type=None, bin_int_type=None, bool_type=None,
                 complex_type=None, date_type=None, datetime_type=None,
                 comment_type=None, comment_id_type=None,
                 float_type=None, grid_type=None, hex_int_type=None,
                 int_type=None, oct_int_type=None, str_type=None, time_type=None,

                 bin_types=None, bin_int_types=None, bool_types=None,
                 complex_types=None, date_types=None, datetime_types=None,
                 comment_types=None, comment_id_types=None,
                 float_types=None, grid_types=None, hex_int_types=None,
                 int_types=None, oct_int_types=None, str_types=None, time_types=None):

        # adapters
        self._adapters = adapters if adapters else dict()

        # containers for deserialization
        self._dict_type = dict_type if dict_type else dict
        self._list_type = list_type if list_type else list
        self._set_type = set_type if set_type else set
        self._obj_type = obj_type if obj_type else box.Obj

        # datatypes for deserialization
        self._bin_type = bin_type if bin_type else bytes
        self._bin_int_type = bin_int_type if bin_int_type else box.BinInt
        self._bool_type = bool_type if bool_type else bool
        self._complex_type = complex_type if complex_type else complex
        self._date_type = date_type if date_type else datetime.date
        self._datetime_type = datetime_type if datetime_type else datetime.datetime
        self._comment_type = comment_type if comment_type else box.Comment
        self._comment_id_type = comment_id_type if comment_id_type else box.CommentID
        self._float_type = float_type if float_type else float
        self._grid_type = grid_type if grid_type else box.Grid
        self._hex_int_type = hex_int_type if hex_int_type else box.HexInt
        self._int_type = int_type if int_type else int
        self._oct_int_type = oct_int_type if oct_int_type else box.OctInt
        self._str_type = str_type if str_type else str
        self._time_type = time_type if time_type else datetime.time

        # serializable containers
        self._dict_types = dict_types if dict_types else [dict, OrderedDict]
        self._list_types = list_types if list_types else [list, tuple]
        self._set_types = set_types if set_types else [set]
        self._obj_types = obj_types if obj_types else [box.Obj]

        # serializable data types
        self._bin_types = bin_types if bin_types else [bytes]
        self._bin_int_types = bin_int_types if bin_int_types else [box.BinInt]
        self._bool_types = bool_types if bool_types else [bool]
        self._complex_types = complex_types if complex_types else [complex]
        self._date_types = date_types if date_types else [datetime.date]
        self._datetime_types = datetime_types if datetime_types else [datetime.datetime]
        self._comment_types = comment_types if comment_types \
            else [box.Comment]
        self._comment_id_types = comment_id_types if comment_id_types \
            else [box.CommentID]
        self._float_types = float_types if float_types else [float, Decimal]
        self._grid_types = grid_types if grid_types else [box.Grid]
        self._hex_int_types = hex_int_types if hex_int_types else [box.HexInt]
        self._int_types = int_types if int_types else [int]
        self._oct_int_types = oct_int_types if oct_int_types else [box.OctInt]
        self._str_types = str_types if str_types else [str]
        self._time_types = time_types if time_types else [datetime.time]

        # setup
        self._mapping = self._create_map()

    # adapters
    @property
    def adapters(self):
        return self._adapters

    @adapters.setter
    def adapters(self, val):
        self._adapters = val

    # types for containers

    @property
    def dict_type(self):
        return self._dict_type

    @dict_type.setter
    def dict_type(self, val):
        self._dict_type = val
        self._update_types("dict", val)

    @property
    def list_type(self):
        return self._list_type

    @list_type.setter
    def list_type(self, val):
        self._list_type = val
        self._update_types("list", val)

    @property
    def set_type(self):
        return self._set_type

    @set_type.setter
    def set_type(self, val):
        self._set_type = val
        self._update_types("set", val)

    @property
    def obj_type(self):
        return self._obj_type

    @obj_type.setter
    def obj_type(self, val):
        self._obj_type = val
        self._update_types("obj", val)

    @property
    def dict_types(self):
        return self._dict_types

    @dict_types.setter
    def dict_types(self, val):
        self._dict_types = val

    @property
    def list_types(self):
        return self._list_types

    @list_types.setter
    def list_types(self, val):
        self._list_types = val

    @property
    def set_types(self):
        return self._set_types

    @set_types.setter
    def set_types(self, val):
        self._set_types = val

    @property
    def obj_types(self):
        return self._obj_types

    @obj_types.setter
    def obj_types(self, val):
        self._obj_types = val

    # types for values

    @property
    def bin_type(self):
        return self._bin_type

    @bin_type.setter
    def bin_type(self, val):
        self._bin_type = val
        self._update_types("bin", val)

    @property
    def bin_int_type(self):
        return self._bin_int_type

    @bin_int_type.setter
    def bin_int_type(self, val):
        self._bin_int_type = val
        self._update_types("bin_int", val)

    @property
    def bool_type(self):
        return self._bool_type

    @bool_type.setter
    def bool_type(self, val):
        self._bool_type = val
        self._update_types("bool", val)

    @property
    def complex_type(self):
        return self._complex_type

    @complex_type.setter
    def complex_type(self, val):
        self._complex_type = val
        self._update_types("complex", val)

    @property
    def date_type(self):
        return self._date_type

    @date_type.setter
    def date_type(self, val):
        self._date_type = val
        self._update_types("date", val)

    @property
    def datetime_type(self):
        return self._datetime_type

    @datetime_type.setter
    def datetime_type(self, val):
        self._datetime_type = val
        self._update_types("datetime", val)

    @property
    def comment_type(self):
        return self._comment_type

    @comment_type.setter
    def comment_type(self, val):
        self._comment_type = val
        self._update_types("comment", val)

    @property
    def comment_id_type(self):
        return self._comment_id_type

    @comment_id_type.setter
    def comment_id_type(self, val):
        self._comment_id_type = val
        self._update_types("comment_id", val)

    @property
    def float_type(self):
        return self._float_type

    @float_type.setter
    def float_type(self, val):
        self._float_type = val
        self._update_types("float", val)

    @property
    def grid_type(self):
        return self._grid_type

    @grid_type.setter
    def grid_type(self, val):
        self._grid_type = val
        self._update_types("grid", val)

    @property
    def hex_int_type(self):
        return self._hex_int_type

    @hex_int_type.setter
    def hex_int_type(self, val):
        self._hex_int_type = val
        self._update_types("hex_int", val)

    @property
    def int_type(self):
        return self._int_type

    @int_type.setter
    def int_type(self, val):
        self._int_type = val
        self._update_types("int", val)

    @property
    def oct_int_type(self):
        return self._oct_int_type

    @oct_int_type.setter
    def oct_int_type(self, val):
        self._oct_int_type = val
        self._update_types("oct_int", val)

    @property
    def str_type(self):
        return self._str_type

    @str_type.setter
    def str_type(self, val):
        self._str_type = val
        self._update_types("str", val)

    @property
    def time_type(self):
        return self._time_type

    @time_type.setter
    def time_type(self, val):
        self._time_type = val
        self._update_types("time", val)

    @property
    def bin_types(self):
        return self._bin_types

    @bin_types.setter
    def bin_types(self, val):
        self._bin_types = val

    @property
    def bin_int_types(self):
        return self._bin_int_types

    @bin_int_types.setter
    def bin_int_types(self, val):
        self._bin_int_types = val

    @property
    def bool_types(self):
        return self._bool_types

    @bool_types.setter
    def bool_types(self, val):
        self._bool_types = val

    @property
    def complex_types(self):
        return self._complex_types

    @complex_types.setter
    def complex_types(self, val):
        self._complex_types = val

    @property
    def date_types(self):
        return self._date_types

    @date_types.setter
    def date_types(self, val):
        self._date_types = val

    @property
    def datetime_types(self):
        return self._datetime_types

    @datetime_types.setter
    def datetime_types(self, val):
        self._datetime_types = val

    @property
    def comment_types(self):
        return self._comment_types

    @comment_types.setter
    def comment_types(self, val):
        self._comment_types = val

    @property
    def comment_id_types(self):
        return self._comment_id_types

    @comment_id_types.setter
    def comment_id_types(self, val):
        self._comment_id_types = val

    @property
    def float_types(self):
        return self._float_types

    @float_types.setter
    def float_types(self, val):
        self._float_types = val

    @property
    def grid_types(self):
        return self._grid_types

    @grid_types.setter
    def grid_types(self, val):
        self._grid_types = val

    @property
    def hex_int_types(self):
        return self._hex_int_types

    @hex_int_types.setter
    def hex_int_types(self, val):
        self._hex_int_types = val

    @property
    def int_types(self):
        return self._int_types

    @int_types.setter
    def int_types(self, val):
        self._int_types = val

    @property
    def oct_int_types(self):
        return self._oct_int_types

    @oct_int_types.setter
    def oct_int_types(self, val):
        self._oct_int_types = val

    @property
    def str_types(self):
        return self._str_types

    @str_types.setter
    def str_types(self, val):
        self._str_types = val

    @property
    def time_types(self):
        return self._time_types

    @time_types.setter
    def time_types(self, val):
        self._time_types = val

    def adapt(self, data):
        """Checks the 'adapters' attribute to find out if there is
        an adapter function registered for the type of the data argument.
        Then, calls the adapter on the data.

        [parameters]
        - data: the data to adapt is an arbitrary Python object

        [return]
        Returns the adapted data or the same data if no adapter is registered for its type"""
        if not self._adapters:
            return data
        t = type(data)
        try:
            return self._adapters[t](data)
        except KeyError as e:
            return data

    def check(self, datatype):
        """This function accepts as argument a Python type, and return
        its Paradict string type if it exists, else returns None"""
        try:
            return self._mapping[datatype]
        except KeyError as e:
            pass

    def _create_map(self):
        categories = {"dict": self._dict_types, "list": self._list_types,
                      "set": self._set_types, "obj": self._obj_types,

                      "bin": self._bin_types, "bin_int": self._bin_int_types,
                      "bool": self._bool_types, "complex": self._complex_types,
                      "date": self._date_types, "datetime": self._datetime_types,
                      "comment": self._comment_types, "comment_id": self._comment_id_types,
                      "float": self._float_types, "grid": self._grid_types,
                      "hex_int": self._hex_int_types, "int": self._int_types,
                      "oct_int": self._oct_int_types, "str": self._str_types,
                      "time": self._time_types}
        mapping = dict()
        for name, datatypes in categories.items():
            for datatype in datatypes:
                mapping[datatype] = name
        return mapping

    def _update_types(self, name, datatype):
        if name == "dict" and datatype not in self._dict_types:
            self._dict_types.insert(0, datatype)
        elif name == "list" and datatype not in self._list_types:
            self._list_types.insert(0, datatype)
        elif name == "set" and datatype not in self._set_types:
            self._set_types.insert(0, datatype)
        elif name == "obj" and datatype not in self._obj_types:
            self._obj_types.insert(0, datatype)
        elif name == "bin" and datatype not in self._bin_types:
            self._bin_types.insert(0, datatype)
        elif name == "bin_int" and datatype not in self._bin_int_types:
            self._bin_int_types.insert(0, datatype)
        elif name == "bool" and datatype not in self._bool_types:
            self._bool_types.insert(0, datatype)
        elif name == "complex" and datatype not in self._complex_types:
            self._complex_types.insert(0, datatype)
        elif name == "date" and datatype not in self._date_types:
            self._date_types.insert(0, datatype)
        elif name == "datetime" and datatype not in self._datetime_types:
            self._datetime_types.insert(0, datatype)
        elif name == "comment" and datatype not in self._comment_types:
            self._comment_types.insert(0, datatype)
        elif name == "comment_id" and datatype not in self._comment_id_types:
            self._comment_id_types.insert(0, datatype)
        elif name == "float" and datatype not in self._float_types:
            self._float_types.insert(0, datatype)
        elif name == "grid" and datatype not in self._grid_types:
            self._grid_types.insert(0, datatype)
        elif name == "hex_int" and datatype not in self._hex_int_types:
            self._hex_int_types.insert(0, datatype)
        elif name == "int" and datatype not in self._int_types:
            self._int_types.insert(0, datatype)
        elif name == "oct_int" and datatype not in self._oct_int_types:
            self._oct_int_types.insert(0, datatype)
        elif name == "str" and datatype not in self._str_types:
            self._str_types.insert(0, datatype)
        elif name == "time" and datatype not in self._time_types:
            self._time_types.insert(0, datatype)
        else:
            msg = "Unknown datatype named '{}'".format(name)
            raise errors.Error(msg)



import unittest
import datetime
from paradict import serializer, deserializer
from paradict import box, errors, tags
from paradict.deserializer.unpacker import Unpacker
from paradict.serializer.packer import Packer


class TestEmptyData(unittest.TestCase):

    def test(self):
        data = dict()
        r = pack_unpack(data)
        self.assertEqual(data, r)


class TestNopTag(unittest.TestCase):

    def test(self):
        unpacker = Unpacker()
        unpacker.feed(tags.NOP)
        self.assertEqual(None, unpacker.data)


class TestDict(unittest.TestCase):

    def test_empty_dict(self):
        d = {0: dict()}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_nested_dict(self):
        val = {0: {1: {2: 3}}}
        d = {0: dict(val)}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIs(type(r[0]), dict)

    def test_dict_filled_with_complex_data(self):
        d = {0: {"age": 999, "name": "John Doe", "pi": 3.14,
                 0: [1, 2, 3], 1: {"a", "b"}, 2000: {"x": 11, "y": dict()},
                 3.14: None, datetime.datetime(2020, 1, 1): True,
                 False: None, None: 3, "misc": [dict(), box.Obj()],
                 "matryoshka": dict(box.Obj(dict(box.Obj(box.Obj()))))}}
        r = pack_unpack(d)
        self.assertEqual(d, r)


class TestList(unittest.TestCase):

    def test_empty_list(self):
        d = {0: list()}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_list_filled_with_complex_data(self):
        d = {0: ["age", 999,  3.14, 0, [1, 2, 3, [4, [5, 6]]], 1, {"a", "b"},
                 2000, {"x": 11}, datetime.datetime(2020, 1, 1),
                 True, False, None]}
        r = pack_unpack(d)
        self.assertEqual(d, r)


class TestSet(unittest.TestCase):

    def test_empty_set(self):
        d = {0: set()}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_set_filled_with_complex_data(self):
        d = {0: {"age", 999, 3.14, 0, 1,
                 2000, datetime.datetime(2020, 1, 1),
                 True, False, None}}
        r = pack_unpack(d)
        self.assertEqual(d, r)


class TestObj(unittest.TestCase):

    def test_empty_obj(self):
        d = {0: box.Obj()}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIs(type(r[0]), box.Obj)

    def test_nested_obj(self):
        val = {1: box.Obj({2: box.Obj({3: box.Obj({4: 5})})})}
        d = {0: box.Obj(val)}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIs(type(r[0]), box.Obj)

    def test_obj_filled_with_complex_data(self):
        val = {"age": 999, "name": "John Doe", "pi": 3.14,
               0: [1, 2, 3], 1: {"a", "b"}, 2000: {"x": 11},
               3.14: None, datetime.datetime(2020, 1, 1): True,
               False: None, None: 3}
        d = {0: box.Obj(val)}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIs(type(r[0]), box.Obj)


class TestObjBuilder(unittest.TestCase):

    def test_simple_obj(self):
        real, imag = -3.14, 4.2
        complex_obj = box.Obj()
        complex_obj["real"] = real
        complex_obj["imag"] = imag
        d = {0: complex_obj}
        packer = Packer()
        unpacker = Unpacker(obj_builder=lambda obj:
                            complex(obj["real"], obj["imag"]))
        for datum in packer.pack(d):
            unpacker.feed(datum)
        expected = {0: complex(real, imag)}
        self.assertEqual(expected, unpacker.data)

    def test_nested_obj(self):
        real, imag = -3.14, 4.2
        # real obj
        real_obj = box.Obj()
        real_obj["type"] = "number"
        real_obj["x"] = real
        # imag obj
        imag_obj = box.Obj()
        imag_obj["type"] = "number"
        imag_obj["x"] = imag
        # complex obj
        complex_obj = box.Obj()
        complex_obj["type"] = "complex"
        complex_obj["real"] = real_obj
        complex_obj["imag"] = imag_obj
        d = {0: complex_obj}
        packer = Packer()
        unpacker = Unpacker(obj_builder=my_obj_builder)
        for datum in packer.pack(d):
            unpacker.feed(datum)
        expected = {0: complex(real, imag)}
        self.assertEqual(expected, unpacker.data)


class TestGrid(unittest.TestCase):

    def test_empty_grid(self):
        d = {0: box.Grid()}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_with_integer(self):
        x = [(0, 1, 2),
             (3, 4, 5_000_000_000),
             (6, 7, -8)]
        d = {0: box.Grid(x)}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_float(self):
        x = [(0.0, 1.1, -2.2_000_000_000)]
        d = {0: box.Grid(x)}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_complex(self):
        x = [(1+2j, 3+4.5_000_000_000j)]
        d = {0: box.Grid(x)}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.Grid)

    def test_with_illegal_data(self):
        x = [("a", "b")]
        d = {0: box.Grid(x)}
        with self.assertRaises(errors.Error):
            pack_unpack(d)


class TestDatetime(unittest.TestCase):

    def test_naive_datetime_lower_endpoint(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.datetime)

    def test_naive_datetime_upper_endpoint(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.datetime)

    def test_aware_datetime_lower_endpoint(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.datetime)

    def test_aware_datetime_upper_endpoint(self):
        year, month = 9999, 12
        day, hour, minute = 31, 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.datetime)

    def test_aware_datetime_with_negative_negative_utc_timezone(self):
        year, month = 1, 1
        day, hour, minute = 1, 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=-23, minutes=-59))
        val = datetime.datetime(year=year, month=month, day=day,
                                hour=hour, minute=minute,
                                second=second, microsecond=microsecond,
                                tzinfo=tz)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.datetime)


class TestDate(unittest.TestCase):

    def test_date_lower_endpoint(self):
        year, month, day = 1, 1, 1
        val = datetime.date(year=year, month=month, day=day)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.date)

    def test_date_upper_endpoint(self):
        year, month, day = 9999, 12, 31
        val = datetime.date(year=year, month=month, day=day)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.date)


class TestTime(unittest.TestCase):

    def test_naive_time_lower_endpoint(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.time)

    def test_naive_time_upper_endpoint(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.time)

    def test_aware_time_lower_endpoint(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.time)

    def test_aware_time_upper_endpoint(self):
        hour, minute = 23, 59
        second, microsecond = 59, 999999
        tz = datetime.timezone(datetime.timedelta(hours=23, minutes=59))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.time)

    def test_aware_time_with_negative_negative_utc_timezone(self):
        hour, minute = 0, 0
        second, microsecond = 0, 0
        tz = datetime.timezone(datetime.timedelta(hours=-23, minutes=-59))
        val = datetime.time(hour=hour, minute=minute,
                            second=second, microsecond=microsecond,
                            tzinfo=tz)
        d = {0: val}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertEqual(type(r[0]), datetime.time)


class TestNull(unittest.TestCase):

    def test(self):
        d = {0: None}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsNone(r[0])


class TestBool(unittest.TestCase):

    def test_true(self):
        d = {0: True}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertTrue(r[0])

    def test_false(self):
        d = {0: False}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertFalse(r[0])


class TestConstString(unittest.TestCase):

    def test_lower_letters(self):
        letters = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j",
                   "k", "l", "m", "n", "o", "p", "q", "r", "s", "t",
                   "u", "v", "w", "x", "y", "z"]
        for x in letters:
            with self.subTest("Letter: {}".format(x)):
                d = {x: x}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_upper_letters(self):
        letters = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J",
                   "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T",
                   "U", "V", "W", "X", "Y", "Z"]
        for x in letters:
            with self.subTest("Letter: {}".format(x)):
                d = {x: x}
                r = pack_unpack(d)
                self.assertEqual(d, r)


class TestString(unittest.TestCase):

    def test_empty_str(self):
        x = str()
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_str_8_to_256(self):
        char = ";"
        for n in range(1, 9):
            with self.subTest("n: {}".format(n)):
                x = char*n
                d = {0: x}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_short_str_lower_endpoint(self):
        x = ";"*33
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_short_str_upper_endpoint(self):
        x = ";"*(2**8)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_medium_str_lower_endpoint(self):
        x = ";" * (2**8 + 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_medium_str_upper_endpoint(self):
        x = ";" * (2**16)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_long_str_lower_endpoint(self):
        x = ";" * (2**16 + 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_long_str_upper(self):
        x = ";" * (2**24)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)

    def test_big_str(self):
        # toooooo big to test ;)
        self.assertTrue(True)

    def test_heavy_str(self):
        # toooooo heavy to test ;)
        self.assertTrue(True)


class TestCommentString(unittest.TestCase):

    def test_empty_comment_str(self):
        x = str()
        d = {box.CommentID(): box.Comment(x)}
        r = pack_unpack(d, skip_comments=False)
        self.assertEqual(d[next(iter(d))], r[next(iter(r))])
        self.assertIsInstance(next(iter(r)), box.CommentID)

    def test_comment_str_8(self):
        x = "#"
        d = {box.CommentID(): box.Comment(x)}
        r = pack_unpack(d, skip_comments=False)
        self.assertEqual(d[next(iter(d))], r[next(iter(r))])
        self.assertIsInstance(next(iter(r)), box.CommentID)

    def test_comment_str_64(self):
        x = "#" * 8
        d = {box.CommentID(): box.Comment(x)}
        r = pack_unpack(d, skip_comments=False)
        self.assertEqual(d[next(iter(d))], r[next(iter(r))])
        self.assertIsInstance(next(iter(r)), box.CommentID)

    def test_short_comment_str(self):
        x = "#" * (2**8)
        d = {box.CommentID(): box.Comment(x)}
        r = pack_unpack(d, skip_comments=False)
        self.assertEqual(d[next(iter(d))], r[next(iter(r))])
        self.assertIsInstance(next(iter(r)), box.CommentID)

    def test_medium_comment_str(self):
        x = "#" * (2**16)
        d = {box.CommentID(): box.Comment(x)}
        r = pack_unpack(d, skip_comments=False)
        self.assertEqual(d[next(iter(d))], r[next(iter(r))])
        self.assertIsInstance(next(iter(r)), box.CommentID)

    def test_long_comment_str(self):
        x = "#" * (2**24)
        d = {box.CommentID(): box.Comment(x)}
        r = pack_unpack(d, skip_comments=False)
        self.assertEqual(d[next(iter(d))], r[next(iter(r))])
        self.assertIsInstance(next(iter(r)), box.CommentID)


class TestBin(unittest.TestCase):

    def test_empty_bin(self):
        x = b''
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_short_bin_lower_endpoint(self):
        x = b'\x00' * 9
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_short_bin_upper_endpoint(self):
        x = b'\x00' * (2**8)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_medium_bin_lower_endpoint(self):
        x = b'\x00' * (2**8 + 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_medium_bin_upper_endpoint(self):
        x = b'\x00' * (2**16)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_long_bin_lower_endpoint(self):
        x = b'\x00' * (2**16 + 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_long_bin_upper(self):
        x = b'\x00' * (2**24)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], bytes)

    def test_big_bin(self):
        # toooooo big to test ;)
        self.assertTrue(True)

    def test_heavy_bin(self):
        # toooooo heavy to test ;)
        self.assertTrue(True)


class TestConstInt(unittest.TestCase):

    def test_const_number_0_to_99(self):
        for x in range(100):
            with self.subTest("Const number: {}".format(x)):
                d = {0: x}
                r = pack_unpack(d)
                self.assertEqual(d, r)


class TestPositiveInt(unittest.TestCase):

    def test_pint8_to_pint64_lower_endpoint(self):
        data = {"pint8": 100, "pint16": 2**8,
                "pint24": 2**16, "pint32": 2**24,
                "pint40": 2**32, "pint48": 2**40,
                "pint56": 2**48, "pint64": 2**56}
        for key, val in data.items():
            with self.subTest("{}".format(key)):
                d = {0: val}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_pint8_to_pint64_upper_endpoint(self):
        data = {"pint8": 2**8 - 1, "pint16": 2**16 - 1,
                "pint24": 2**24 - 1, "pint32": 2**32 - 1,
                "pint40": 2**40 - 1, "pint48": 2**48 - 1,
                "pint56": 2**56 - 1, "pint64": 2**64 - 1}
        for key, val in data.items():
            with self.subTest("{}".format(key)):
                d = {0: val}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_pint_big_lower(self):
        x = 2**69
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)

    def test_pint_big_upper(self):
        x = 2**2048 - 1
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)

    def test_pint_heavy_lower(self):
        x = 2**2048
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)

    def test_pint_heavy_upper(self):
        x = 2**524288 - 1
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)


class TestNegativeInt(unittest.TestCase):

    def test_negative_number_1_to_99(self):
        for x in range(1, 100):
            with self.subTest("Number: {}".format(x)):
                d = {0: -x}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_pint8_to_pint64_lower_endpoint(self):
        data = {"pint8": 100, "pint16": 2**8,
                "pint24": 2**16, "pint32": 2**24,
                "pint40": 2**32, "pint48": 2**40,
                "pint56": 2**48, "pint64": 2**56}
        for key, val in data.items():
            with self.subTest("{}".format(key)):
                d = {0: -val}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_pint8_to_pint64_upper_endpoint(self):
        data = {"pint8": 2**8 - 1, "pint16": 2**16 - 1,
                "pint24": 2**24 - 1, "pint32": 2**32 - 1,
                "pint40": 2**40 - 1, "pint48": 2**48 - 1,
                "pint56": 2**56 - 1, "pint64": 2**64 - 1}
        for key, val in data.items():
            with self.subTest("{}".format(key)):
                d = {0: -val}
                r = pack_unpack(d)
                self.assertEqual(d, r)

    def test_nint_big_lower(self):
        x = 2**69
        data = {0: -x}
        r = pack_unpack(data)
        self.assertEqual(data, r)

    def test_nint_big_upper(self):
        x = 2**2048 - 1
        data = {0: -x}
        r = pack_unpack(data)
        self.assertEqual(data, r)

    def test_nint_heavy_lower(self):
        x = 2**2048
        data = {0: -x}
        r = pack_unpack(data)
        self.assertEqual(data, r)

    def test_nint_heavy_upper(self):
        x = 2**524288 - 1
        data = {0: -x}
        r = pack_unpack(data)
        self.assertEqual(data, r)


class TestIntAsHexOctBin(unittest.TestCase):

    def test_zero_as_hex(self):
        x = box.HexInt(0)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.HexInt)

    def test_zero_as_oct(self):
        x = box.OctInt(0)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.OctInt)

    def test_zero_as_bin(self):
        x = box.BinInt(0)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.BinInt)

    def test_pint_as_hex(self):
        x = box.HexInt(2**8 - 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.HexInt)

    def test_pint_as_oct(self):
        x = box.OctInt(2**8 - 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.OctInt)

    def test_pint_as_bin(self):
        x = box.BinInt(2**8 - 1)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.BinInt)

    def test_nint_as_hex(self):
        val = 2**8 - 1
        x = box.HexInt(-val)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.HexInt)

    def test_nint_as_oct(self):
        x = 2**8 - 1
        x = box.OctInt(-x)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.OctInt)

    def test_nint_as_bin(self):
        val = 2**8 - 1
        x = box.BinInt(-val)
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.BinInt)


class TestIntAsStringHexOctBin(unittest.TestCase):

    def test_zero_as_hex(self):
        x = box.HexInt("0x0000_0000")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.HexInt)

    def test_zero_as_oct(self):
        x = box.OctInt("0o000_000")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.OctInt)

    def test_zero_as_bin(self):
        x = box.BinInt("0b0000_0000")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.BinInt)

    def test_pint_as_hex(self):
        x = box.HexInt("0x0000_00ff")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.HexInt)

    def test_pint_as_oct(self):
        x = box.OctInt("0o000_377")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.OctInt)

    def test_pint_as_bin(self):
        x = box.BinInt("0b0000_1111_1111")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.BinInt)

    def test_nint_as_hex(self):
        x = box.HexInt("-0x0000_00ff")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.HexInt)

    def test_nint_as_oct(self):
        x = box.OctInt("-0o000_377")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.OctInt)

    def test_nint_as_bin(self):
        x = box.BinInt("-0b0000_1111_1111")
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], box.BinInt)


class TestFloatNumber(unittest.TestCase):

    def test_zero(self):
        x = 0.0
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_negative_zero(self):
        x = float("-0.0")
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_positive_infinity(self):
        x = float("+inf")
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_negative_infinity(self):
        x = float("-inf")
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_positive_number_with_exponent(self):
        x = float("3.0{}E-10".format(2**32 - 1))
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_positive_number_without_exponent(self):
        x = float("3.00{}".format(2**32 - 1))
        data = {0: x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_negative_number_with_exponent(self):
        x = float("3.0{}E-10".format(2**32 - 1))
        data = {0: -x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)

    def test_negative_number_without_exponent(self):
        x = float("3.{}00".format(2**32 - 1))
        data = {0: -x}
        r = pack_unpack(data)
        self.assertEqual(data, r)
        self.assertIsInstance(r[0], float)


class TestComplexNumber(unittest.TestCase):

    def test_with_positive_number(self):
        x = 1.2+3.4j
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], complex)

    def test_with_negative_number(self):
        x = -1.2-3.4j
        d = {0: x}
        r = pack_unpack(d)
        self.assertEqual(d, r)
        self.assertIsInstance(r[0], complex)


def pack_unpack(data, skip_comments=True):
    data = serializer.pack(data, skip_comments=skip_comments)
    return deserializer.unpack(data, skip_comments=skip_comments)


def my_obj_builder(obj):
    datatype = obj["type"]
    if datatype == "number":
        return obj["x"]
    elif datatype == "complex":
        return complex(obj["real"], obj["imag"])


if __name__ == "__main__":
    unittest.main()



import unittest
from paradict import box


class TestHexInt(unittest.TestCase):

    def test_with_int(self):
        for key, val in {0: 0,
                         1: 1,
                         -255: -255}.items():
            with self.subTest("Val: {}".format(key)):
                r = box.HexInt(key)
                self.assertEqual(val, r)

    def test_with_str_int(self):
        for key, val in {"0": 0,
                         "1": 1,
                         "-255": -255,
                         "0x0": 0,
                         "0x1": 1,
                         "-0xff": -255}.items():
            with self.subTest("Val: {}".format(key)):
                r = box.HexInt(key)
                self.assertEqual(val, r)

    def test_string_output(self):
        for key, val in {0: "0x0",
                         1: "0x1",
                         -255: "-0xff",
                         "0x0": "0x0",
                         "0x1": "0x1",
                         "-0x000000ff": "-0x0000_00ff"}.items():
            with self.subTest("Val: {}".format(key)):
                r = str(box.HexInt(key)).lower()
                self.assertEqual(val, r)

    def test_with_different_base(self):
        r = box.HexInt("-0b11111111")
        self.assertEqual(-255, r)

    def test_string_output_with_different_base(self):
        r = box.HexInt("-0b1111_1111")  # the number of digits counts
        self.assertEqual("-0xff", str(r).lower())


class TestOctInt(unittest.TestCase):

    def test_with_int(self):
        for key, val in {0: 0,
                         1: 1,
                         -255: -255}.items():
            with self.subTest("Val: {}".format(key)):
                r = box.OctInt(key)
                self.assertEqual(val, r)

    def test_with_str_int(self):
        for key, val in {"0": 0,
                         "1": 1,
                         "-255": -255,
                         "0o0": 0,
                         "0o1": 1,
                         "-0o377": -255}.items():
            with self.subTest("Val: {}".format(key)):
                r = box.OctInt(key)
                self.assertEqual(val, r)

    def test_string_output(self):
        for key, val in {0: "0o0",
                         1: "0o1",
                         -255: "-0o377",
                         "0o0": "0o0",
                         "0o1": "0o1",
                         "-0o000377": "-0o000_377"}.items():
            with self.subTest("Val: {}".format(key)):
                r = str(box.OctInt(key)).lower()
                self.assertEqual(val, r)

    def test_with_different_base(self):
        r = box.OctInt("-0xff")
        self.assertEqual(-255, r)

    def test_string_output_with_different_base(self):
        r = box.OctInt("-0x00ff")
        self.assertEqual("-0o377", str(r).lower())


class TestBinInt(unittest.TestCase):

    def test_with_int(self):
        for key, val in {0: 0,
                         1: 1,
                         -255: -255}.items():
            with self.subTest("Val: {}".format(key)):
                r = box.BinInt(key)
                self.assertEqual(val, r)

    def test_with_str_int(self):
        for key, val in {"0": 0,
                         "1": 1,
                         "-255": -255,
                         "0b0": 0,
                         "0b1": 1,
                         "-0b11111111": -255}.items():
            with self.subTest("Val: {}".format(key)):
                r = box.BinInt(key)
                self.assertEqual(val, r)

    def test_string_output(self):
        for key, val in {0: "0b0",
                         1: "0b1",
                         -255: "-0b1111_1111",
                         "0b0": "0b0",
                         "0b1": "0b1",
                         "-0b000011111111": "-0b0000_1111_1111"}.items():
            with self.subTest("Val: {}".format(key)):
                r = str(box.BinInt(key)).lower()
                self.assertEqual(val, r)

    def test_with_different_base(self):
        r = box.BinInt("-0xff")
        self.assertEqual(-255, r)

    def test_string_output_with_different_base(self):
        r = box.BinInt("-0xff")
        self.assertEqual("-0b1111_1111", str(r).lower())


if __name__ == '__main__':
    unittest.main()



"""High-level functions to serialize Python dict in Paradict binary/text format"""
import written
from paradict.serializer.encoder import Encoder
from paradict.serializer.packer import Packer
from paradict import const


def encode(data, *, mode=const.DATA_MODE, type_ref=None,
           skip_comments=False, skip_bin_data=False):
    """
    Convert a Python dictionary object to Paradict binary format

    [parameters]
    - data: Python dict
    - mode: either const.DATA_MODE or const.CONFIG_MODE. Defaults to DATA_MODE.
    - type_ref: optional TypeRef object
    - skip_comments: boolean to tell whether comments should be ignored or not
    - skip_bin_data: boolean to tell whether bin data should be ignored or not

    [return]
    Return a string in the Paradict text format
    """
    encoder = Encoder(mode=mode, type_ref=type_ref,
                      skip_comments=skip_comments,
                      skip_bin_data=skip_bin_data)
    lines = list()
    for r in encoder.encode(data):
        lines.append(r)
    return "\n".join(lines)


def pack(data, *, type_ref=None, skip_comments=False):
    """
    Convert a Python dictionary object to Paradict binary format

    [parameters]
    - data: Python dict
    - type_ref: optional TypeRef object
    - skip_comments: boolean to tell whether comments should be ignored or not

    [return]
    Return a Python bytes object packed in the Paradict binary format
    """
    packer = Packer(type_ref=type_ref, skip_comments=skip_comments)
    buffer = bytearray()
    for r in packer.pack(data):
        buffer.extend(r)
    return bytes(buffer)


def dump(data, path, *, type_ref=None, skip_comments=False):
    """
    Convert some Python dict in the Paradict binary format
    then dump it in a file
    [parameters]
    - path: path string or pathlib.Path instance
    - type_ref: optional TypeRef object
    - skip_comments: boolean to tell whether comments should be ignored or not
    """
    r = pack(data, skip_comments=skip_comments, type_ref=type_ref)
    written.write(r, path)
